# -*- coding: utf-8 -*-

from __future__ import annotations

import typing
import warnings
import datetime

import numpy as np
import pandas as pd
import scipy.signal

from endaq.calc import utils


def _np_histogram_nd(array, bins=10, weights=None, axis=-1, **kwargs):
    """Compute histograms over a specified axis."""
    array = np.asarray(array)
    bins = np.asarray(bins)
    weights = np.asarray(weights) if weights is not None else None

    # Collect all ND inputs
    nd_params = {}
    (nd_params if array.ndim > 1 else kwargs)["a"] = array
    (nd_params if bins.ndim > 1 else kwargs)["bins"] = bins
    if weights is not None:
        (nd_params if weights.ndim > 1 else kwargs)["weights"] = weights

    if len(nd_params) == 0:
        return np.histogram(**kwargs)[0]

    # Move the desired axes to the back
    for k, v in nd_params.items():
        nd_params[k] = np.moveaxis(v, axis, -1)

    # Broadcast ND arrays to the same shape
    ks, vs = zip(*nd_params.items())
    vs_broadcasted = np.broadcast_arrays(*vs)
    for k, v in zip(ks, vs_broadcasted):
        nd_params[k] = v

    # Generate output
    bins = nd_params.get("bins", bins)

    result_shape = ()
    if len(nd_params) != 0:
        result_shape = v.shape[:-1]
    if bins.ndim >= 1:
        result_shape = result_shape + (bins.shape[-1] - 1,)
    else:
        result_shape = result_shape + (bins,)

    result = np.empty(
        result_shape,
        dtype=(weights.dtype if weights is not None else int),
    )
    loop_shape = v.shape[:-1]

    for nd_i in np.ndindex(*loop_shape):
        nd_kwargs = {k: v[nd_i] for k, v in nd_params.items()}
        result[nd_i], edges = np.histogram(**nd_kwargs, **kwargs)

    result = np.moveaxis(result, -1, axis)
    return result


def welch(
    df: pd.DataFrame,
    bin_width: float = 1.0,
    scaling: typing.Literal[None, "density", "spectrum", "parseval"] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Perform `scipy.signal.welch` with a specified frequency spacing.

    :param df: the input data
    :param bin_width: the desired width of the resulting frequency bins, in Hz;
        defaults to 1 Hz
    :param scaling: the scaling of the output; `"density"` & `"spectrum"`
        correspond to the same options in ``scipy.signal.welch``; `"parseval"`
        will maintain the "energy" between the input & output, s.t.
        ``welch(df, scaling="parseval").sum(axis="rows")`` is roughly equal to
        ``df.abs().pow(2).sum(axis="rows")``;
        `"unit"` will maintain the units and scale of the input data.
    :param kwargs: other parameters to pass directly to ``scipy.signal.welch``
    :return: a periodogram

    .. seealso::

        - `SciPy Welch's method <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_
          Documentation for the periodogram function wrapped internally.
        - `Parseval's Theorem <https://en.wikipedia.org/wiki/Parseval's_theorem>`_
          - the theorem relating the RMS of a time-domain signal to that of its
          frequency spectrum
    """
    dt = utils.sample_spacing(df)
    fs = 1 / dt

    if scaling == "parseval":
        kwargs["scaling"] = "density"
    elif scaling == "unit":
        kwargs["scaling"] = "density"
    elif scaling is not None:
        kwargs["scaling"] = scaling

    nperseg = int(fs / bin_width)
    if "nfft" not in kwargs:
        nfft = nperseg
    else:
        nfft = kwargs["nfft"]

    freqs, psd = scipy.signal.welch(df.values, fs=fs, nperseg=nperseg, **kwargs, axis=0)
    if scaling == "parseval":
        psd = psd * freqs[1]
    elif scaling == "unit":
        psd *= 2 * freqs[1]
        psd *= nfft / nperseg
        psd **= 0.5

    return pd.DataFrame(
        psd, index=pd.Series(freqs, name="frequency (Hz)"), columns=df.columns
    )


def differentiate(df: pd.DataFrame, n: float = 1.0) -> pd.DataFrame:
    """
    Perform time-domain differentiation on periodogram data.

    :param df: a periodogram
    :param n: the time derivative order; negative orders represent integration
    :return: a periodogram of the time-differentiated data
    """
    # Involves a division by zero for n < 0
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        factor = (2 * np.pi * df.index.values) ** (2 * n)  # divide by zero
    if n < 0:
        factor[df.index == 0] = 0

    return df * factor[..., np.newaxis]


def to_jagged(
    df: pd.DataFrame,
    freq_splits: np.array,
    agg: typing.Union[
        typing.Literal["mean", "sum"],
        typing.Callable[[np.ndarray, int], float],
    ] = "mean",
) -> pd.DataFrame:
    """
    Calculate a periodogram over non-uniformly spaced frequency bins.

    :param df: the returned values from :py:func:`endaq.calc.psd.welch`
    :param freq_splits: the boundaries of the frequency bins; must be strictly
        increasing
    :param agg: the method for aggregating values into bins; `'mean'` preserves
        the PSD's area-under-the-curve, `'sum'` preserves the PSD's "energy"
    :return: a periodogram with the given frequency spacing
    """
    freq_splits = np.asarray(freq_splits)
    if len(freq_splits) < 2:
        raise ValueError("need at least two frequency bounds")
    if not np.all(freq_splits[:-1] <= freq_splits[1:]):
        raise ValueError("frequency bounds must be strictly increasing")

    # Check that PSD samples do not skip any frequency bins
    spacing_test = np.diff(np.searchsorted(freq_splits, df.index))
    if np.any(spacing_test < 1):
        warnings.warn(
            "empty frequency bins in re-binned PSD; "
            "original PSD's frequency spacing is too coarse",
            RuntimeWarning,
        )

    if isinstance(agg, str):
        if agg not in ("sum", "mean"):
            raise ValueError(f'invalid aggregation mode "{agg}"')

        # Reshape frequencies for histogram function
        f_ndim = np.broadcast_to(df.index.to_numpy()[..., np.newaxis], df.shape)

        # Calculate sum via histogram function
        psd_jagged = _np_histogram_nd(f_ndim, bins=freq_splits, weights=df, axis=0)

        # Adjust values for mean calculation
        if agg == "mean":
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)  # x/0

                psd_jagged = np.nan_to_num(  # <- fix divisions by zero
                    psd_jagged
                    / np.histogram(df.index, bins=freq_splits)[0][..., np.newaxis]
                )

    else:
        psd_binned = np.split(df, np.searchsorted(df.index, freq_splits), axis=0)[1:-1]
        psd_jagged = np.stack([agg(a, axis=0) for a in psd_binned], axis=0)

    return pd.DataFrame(
        psd_jagged,
        index=pd.Series((freq_splits[1:] + freq_splits[:-1]) / 2, name=df.index.name),
        columns=df.columns,
    )


def to_octave(
    df: pd.DataFrame, fstart: float = 1.0, octave_bins: float = 12.0, **kwargs
) -> pd.DataFrame:
    """
    Calculate a periodogram over log-spaced frequency bins.

    :param df: the returned values from :py:func:`endaq.calc.psd.welch`
    :param fstart: the first frequency bin, in Hz; defaults to 1 Hz
    :param octave_bins: the number of frequency bins in each octave; defaults
        to 12
    :param kwargs: other parameters to pass directly to :py:func:`to_jagged`
    :return: a periodogram with the given logarithmic frequency spacing
    """
    max_f = df.index.max()

    octave_step = 1 / octave_bins
    center_freqs = 2 ** np.arange(
        np.log2(fstart),
        np.log2(max_f) + octave_step / 2,
        octave_step,
    )
    freq_splits = 2 ** np.arange(
        np.log2(fstart) - octave_step / 2,
        np.log2(max_f) + octave_step,
        octave_step,
    )
    assert len(center_freqs) + 1 == len(freq_splits)

    result = to_jagged(df, freq_splits=freq_splits, **kwargs)
    result.index = pd.Series(center_freqs, name=df.index.name)
    return result


def _aligned_bin_width(fstart: float = 1.0, octave_bins: float = 12.0) -> float:
    """
    Calculate a "good" linearly-spaced bin width, from a PSD of which an
    octave-spaced PSD can be accurately calculated.

    When computing the octave-spaced PSD, linear bins are mapped to octave bins
    and the linear bin values are then summed together within each octave bin.
    However, this calculation can produce an octave-spaced PSD that is an
    *inaccurate* representation of the original linearly-spaced PSD, if the
    mapping from linear to octave bins is *poorly aligned*; e.g., if a
    significant linear bin is halfway between two adjacent octave bins.

    Thus, the main criterion for a "good" linear spacing is how well the bin
    borders in the linearly-spaced PSD align with those of the octave-spaced
    PSD. This is mostly only significant at the lower frequencies, where the
    octave-spaced bins comprise only a handful of linearly-spaced bins.

    :param fstart: the starting frequency of the octave-spaced PSD
    :param octave_bins: the number of bins per octave in the octave-spaced PSD
    :return: a bin-width for a linearly-spaced PSD that aligns well with an
        octave-spaced PSD defined by the given spacing parameters

    .. todo::
        allow user to adjust output resolution:
        - add "good"-ness as a parameter to adjust resolution of result?
        - iterate over solutions with increasing resolutions -> user just picks
          the most appropriate one?
    """
    # To get a good alignment with an octave-spaced frequency bin, the linear
    # bin width should:
    # - be a *"half" divisor* of the octave bin's lower and upper bounds:
    #       f_oct_lower = f_lin * (n1 + 1/2)
    #       f_oct_upper = f_lin * (n2 + 1/2)
    #   for integers n1, n2
    # -> be a divisor of the octave bin's breadth:
    #       f_oct_upper - f_oct_lower = f_lin * (n2 - n1) = f_lin * m
    lbound_factor = 2 ** (-1 / (2 * octave_bins))
    ubound_factor = 2 ** (1 / (2 * octave_bins))
    breadth_factor = ubound_factor - lbound_factor

    # TODO find another way to calculate a linear bin width that better fits
    # the above description
    arb_odd_int = 5  # any odd integer; 5 provides some resolution w/o too many segments
    return fstart / int(arb_odd_int / breadth_factor)


def vc_curves(
    accel_psd: pd.DataFrame, fstart: float = 1.0, octave_bins: float = 12.0
) -> pd.DataFrame:
    """
    Calculate Vibration Criterion (VC) curves from an acceleration periodogram.

    :param accel_psd: a periodogram of the input acceleration
    :param fstart: the first frequency bin
    :param octave_bins: the number of frequency bins in each octave; defaults
        to 12
    :return: the Vibration Criterion (VC) curve of the input acceleration
    """
    """
    # Theory behind the calculation

    Let x(t) be a real-valued time-domain signal, and X(2πf) = F{x(t)}(2πf)
    be the Fourier Transform of that signal. By Parseval's Theorem,

        ∫x(t)^2 dt = ∫|X(2πf)|^2 df

    (see https://en.wikipedia.org/wiki/Parseval%27s_theorem#Notation_used_in_physics)

    Rewriting the right side of that equation in the discrete form becomes

        ∫x(t)^2 dt ≈ ∑ |X[k]|^2 • ∆f

    where ∆f = fs/N = (1/∆t) / N = 1/T.
    Limiting the right side to a range of discrete frequencies (k_0, k_1):

        ∫x(t)^2 dt ≈ [∑; k=k_0 -> k≤k_1] |X[k]|^2 • ∆f

    The VC curve calculation is the RMS over the time-domain. If T is the
    duration of the time-domain signal, then:

        √((1/T) ∫x(t)^2 dt)
            ≈ √((1/T) [∑; k=k_0 -> k≤k_1] |X[k]|^2 • ∆f)
            = ∆f • √([∑; k=k_0 -> k≤k_1] |X[k]|^2)

    If the time-series data is acceleration, then the signal needs to first
    be integrated into velocity. This can be done in the frequency domain
    by replacing |X(2πf)|^2 with (1/2πf)^2 |X(2πf)|^2.
    """
    df_vel = differentiate(accel_psd, n=-1)
    df_vel_oct = to_octave(
        df_vel,
        fstart=fstart,  # Hz
        octave_bins=octave_bins,
        agg="sum",
    )

    # The PSD must already scale by ∆f -> need only scale by √∆f?
    # TODO make `density` parameter, scale differently depending on mode
    return np.sqrt(accel_psd.index[1]) * df_vel_oct.apply(np.sqrt)


def rolling_psd(
        df: pd.DataFrame,
        bin_width: float = 1.0,
        octave_bins: float = None,
        fstart: float = 1.0,
        scaling: typing.Literal[None, "density", "spectrum", "parseval", "unit", "rms"] = None,
        agg: typing.Union[
            typing.Literal["mean", "sum"],
            typing.Callable[[np.ndarray, int], float],
        ] = "mean",
        freq_splits: np.array = None,
        num_slices: int = 100,
        indexes: np.array = None,
        index_values: np.array = None,
        slice_width: float = None,
        add_resultant: bool = True,
        disable_warnings: bool = True,
        **kwargs,
) -> pd.DataFrame:
    """
    Compute PSDs for defined slices of a time series data set using :py:func:`~endaq.calc.psd.welch()`
    
    :param df: the input dataframe with an index defining the time in seconds or datetime
    :param bin_width: the bin width or resolution in Hz for the PSD, defaults to 1,
        this is ignored if `octave_bins` is defined
    :param octave_bins: the number of frequency bins in each octave, defaults to `None`
    :param fstart: the lowest frequency for an octave PSD, defaults to 1
    :param scaling: the scaling of the output; `"density"` & `"spectrum"`
        correspond to the same options in ``scipy.signal.welch``; `"parseval"`
        will maintain the "energy" between the input & output, s.t.
        ``welch(df, scaling="parseval").sum(axis="rows")`` is roughly equal to
        ``df.abs().pow(2).sum(axis="rows")``;
        `"unit"` will maintain the units and scale of the input data. `"rms"` will use `"parseval"` for the PSD
        calculations and set `agg` to "sum", but then take the square root at the end
    :param agg: the method for aggregating values into bins (only used if converting to octave or jagged); `'mean'` preserves
        the PSD's area-under-the-curve, `'sum'` preserves the PSD's "energy"
    :param freq_splits: the boundaries of the frequency bins to pass to :py:func:`~endaq.calc.psd.to_jagged()`
    :param num_slices: the number of slices to split the time series into, default is 100,
        this is ignored if `indexes` is defined
    :param indexes: the center index locations (not value) of each slice to compute the PSD
    :param index_values: the index values of each peak event to quantify (slower but more intuitive than using `indexes`)
    :param slice_width: the time in seconds to center about each slice index,
        if none is provided it will calculate one based upon the number of slices
    :param add_resultant: if `True` (default) the root sum of squares of each PSD column will
        also be computed
    :param disable_warnings: if `True` (default) it disables the warnings on the PSD length
    :param kwargs: Other parameters to pass directly to :py:func:`psd.welch`
    :return: a dataframe containing all the PSDs, stacked on each other

    See example use cases and syntax at :py:func:`~endaq.plot.spectrum_over_time()`
    which visualizes the output of this function in Heatmaps, Waterfall plots, 
    Surface plots, and Animations

    """
    use_spectrogram = True
    if (indexes is not None) | (index_values is not None):
        use_spectrogram = False

    indexes, slice_width, num, length = utils._rolling_slice_definitions(
        df,
        indexes=indexes,
        index_values=index_values,
        num_slices=num_slices,
        slice_width=slice_width
    )

    if use_spectrogram:
        if "window" not in kwargs:
            kwargs["window"] = "hann"
        psd_df = spectrogram(
            df=df,
            bin_width=bin_width,
            num_slices=num_slices,
            scaling=scaling,
            octave_bins=octave_bins,
            fstart=fstart,
            freq_splits=freq_splits,
            add_resultant=add_resultant,
            disable_warnings=disable_warnings,
            **kwargs
        )
    else:
        # Allow for RMS calculations
        exp = 1
        if scaling == "rms":
            scaling = "parseval"
            exp = 0.5
            agg = "sum"

        # Define bin_width if octave spacing
        if octave_bins is not None:
            bin_width = 1 / slice_width

        # Loop through and compute PSD
        psd_df = pd.DataFrame()
        for i in indexes:
            window_start = max(0, i - num)
            window_end = min(length, i + num)
            with warnings.catch_warnings():
                if disable_warnings:
                    warnings.filterwarnings('ignore', '.*greater than.*', )
                slice_psd = welch(
                    df.iloc[window_start:window_end],
                    bin_width=bin_width,
                    scaling=scaling,
                    **kwargs
                )

            if octave_bins is not None or freq_splits is not None:
                with warnings.catch_warnings():
                    if disable_warnings:
                        warnings.filterwarnings('ignore', '.*empty frequency.*')
                    if freq_splits is None:
                        slice_psd = to_octave(slice_psd, octave_bins=octave_bins, fstart=fstart, agg=agg)
                    else:
                        slice_psd = to_jagged(slice_psd, freq_splits=freq_splits, agg=agg)

            if add_resultant:
                slice_psd['Resultant'] = slice_psd.sum(axis=1)

            slice_psd = (slice_psd ** exp).reset_index().melt(id_vars=slice_psd.index.name)
            slice_psd['timestamp'] = df.index[i]
            psd_df = pd.concat([psd_df, slice_psd])

    return psd_df


def spectrogram(
        df: pd.DataFrame,
        num_slices: int = 100,
        scaling: typing.Literal[None, "density", "spectrum", "parseval", "unit", "rms"] = None,
        bin_width: float = 1.0,
        octave_bins: float = None,
        fstart: float = 1.0,
        agg: typing.Union[
            typing.Literal["mean", "sum"],
            typing.Callable[[np.ndarray, int], float],
        ] = "mean",
        freq_splits: np.array = None,
        add_resultant: bool = True,
        disable_warnings: bool = True,
        **kwargs,
) -> pd.DataFrame:
    """
    Compute a spectrogram for a time series data set using :py:func:`scipy.signal.spectrogram()`

    :param df: the input dataframe with an index defining the time in seconds or datetime
    :param num_slices: the number of slices to split the time series into, default is 100
    :param scaling: the scaling of the output; `"density"` & `"spectrum"`
        correspond to the same options in ``scipy.signal.welch``; `"parseval"`
        will maintain the "energy" between the input & output, s.t.
        ``welch(df, scaling="parseval").sum(axis="rows")`` is roughly equal to
        ``df.abs().pow(2).sum(axis="rows")``;
        `"unit"` will maintain the units and scale of the input data. `"rms"` will use `"parseval"` for the PSD
        calculations and set `agg` to "sum", but then take the square root at the end
    :param bin_width: the bin width or resolution in Hz for the PSD or FFT, defaults to 1
    :param octave_bins: the number of frequency bins in each octave, defaults to `None`
    :param fstart: the lowest frequency for an octave spaced output, defaults to 1
    :param agg: the method for aggregating values into bins (only used if converting to octave or jagged); `'mean'` preserves
        the PSD's area-under-the-curve, `'sum'` preserves the PSD's "energy"
    :param freq_splits: the boundaries of the frequency bins to pass to :py:func:`~endaq.calc.psd.to_jagged()`
    :param add_resultant: if `True` (default) the root sum of squares of each PSD column will
        also be computed
    :param disable_warnings: if `True` (default) it disables the warnings in helper functions
    :param kwargs: Other parameters to pass directly to :py:func:`scipy.signal.spectrogram()`
    :return: a dataframe containing all the spectrograms "melted" with columns defining the value, frequency, timeslice,
        and original column from the input dataframe

    .. seealso::

        - `SciPy Spectrogram method <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.spectrogram.html>`_
          Documentation for the function wrapped internally.
    """

    # Apply Scaling
    if scaling == "parseval":
        kwargs["scaling"] = "density"
    elif scaling == "unit":
        kwargs["scaling"] = "density"
    elif scaling == "rms":
        kwargs["scaling"] = "density"
        agg = "sum"
    elif scaling is not None:
        kwargs["scaling"] = scaling

    # Windowing, Default to Boxcar
    if "window" not in kwargs:
        kwargs["window"] = "boxcar"

    # Determine Scipy.Spectrogram Stats
    fs = 1 / utils.sample_spacing(df)
    nperseg = int(fs / bin_width)
    num_f = int(len(df) / fs / bin_width)
    noverlap = nperseg - (nperseg * num_f // num_slices)

    # Loop Through and Compute Spectrogram
    spec_df = pd.DataFrame()
    if add_resultant:
        res_df = pd.DataFrame()
    for c in df.columns:
        f, t, spec = scipy.signal.spectrogram(
            x=df[c].to_numpy(),
            fs=fs,
            nperseg=nperseg,
            mode='psd',
            noverlap=noverlap,
            **kwargs
        )

        # create dataframe
        spec_df_col = pd.DataFrame(
            data=spec,
            columns=t,
            index=f
        )
        spec_df_col.index.name = 'frequency (Hz)'

        # Do octave spacing
        if octave_bins is not None or freq_splits is not None:
            with warnings.catch_warnings():
                if disable_warnings:
                    warnings.filterwarnings('ignore', '.*empty frequency.*')
                if freq_splits is None:
                    spec_df_col = to_octave(spec_df_col, octave_bins=octave_bins, fstart=fstart, agg=agg)
                else:
                    spec_df_col = to_jagged(spec_df_col, freq_splits=freq_splits, agg=agg)

        # Melt & concatenate dataframe
        spec_df_col = spec_df_col.reset_index().melt(id_vars='frequency (Hz)')
        spec_df_col.columns = ['frequency (Hz)', 'timestamp', 'value']
        spec_df_col['variable'] = c
        spec_df = pd.concat([spec_df, spec_df_col])

        # add resultant
        if add_resultant:
            if len(res_df) == 0:
                res_df = spec_df_col
                res_df['variable'] = "Resultant"
            else:
                res_df.value += spec_df_col.value

    # Add resultant
    if add_resultant:
        spec_df = pd.concat([spec_df, res_df])

    # Update timestamps to be based on input, deal with datetime
    if isinstance(df.index[0], datetime.datetime):
        spec_df.timestamp = pd.to_timedelta(spec_df.timestamp.astype(float), unit='s') + df.index[0]
    else:
        spec_df.timestamp = spec_df.timestamp.astype(float) + df.index[0]

    # Apply scales
    if scaling == "parseval":
        spec_df.value = spec_df.value * f[1]
    if scaling == "rms":
        spec_df.value = spec_df.value ** 0.5
    elif scaling == "unit":
        spec_df.value = (spec_df.value * 2 * f[1]) ** 0.5

    return spec_df



import numpy as np
import pandas as pd
from typing import Optional
import plotly.express as px
import plotly.graph_objects as go


def neg_squared_euc_dists(ary: np.ndarray):
    """
    Compute matrix containing negative squared euclidean distance for all pairs of points in input matrix X
    (Implementation based around the following: https://nlml.github.io/in-raw-numpy/in-raw-numpy-t-sne/)
    (Math? See https://stackoverflow.com/questions/37009647)

    :param ary: A matrix of size NxD
    :return: An NxN matrix D, with entry D_ij = negative squared euclidean distance between rows X_i and
    """
    #
    ary_sum = np.sum(np.square(ary), 1)
    return - np.add(np.add(-2 * np.dot(ary, ary.T), ary_sum).T, ary_sum)


def softmax(ary: np.ndarray, diag_zero: bool = True, zero_index: Optional[int] = None):
    """
    Compute softmax values for each row of matrix X.
    (Implementation based around the following: https://nlml.github.io/in-raw-numpy/in-raw-numpy-t-sne/)
    """

    # Subtract max for numerical stability
    e_x = np.exp(ary - np.max(ary, axis=1).reshape([-1, 1]))

    # We usually want diagonal probabilities to be 0.
    if zero_index is None:
        if diag_zero:
            np.fill_diagonal(e_x, 0.)
    else:
        e_x[:, zero_index] = 0.

    # Add a tiny constant for stability of log we take later
    # e_x = e_x + 1e-8  # numerical stability
    e_x += 1e-8  # for numerical stability

    return e_x / e_x.sum(axis=1).reshape([-1, 1])


def calc_prob_matrix(distances: np.ndarray, sigmas: Optional[np.ndarray] = None, zero_index: Optional[int] = None):
    """
    Convert a distances matrix to a matrix of probabilities.
    (Implementation based around the following: https://nlml.github.io/in-raw-numpy/in-raw-numpy-t-sne/)
    """
    two_sig_sq = 2. * np.square(sigmas.reshape((-1, 1)))
    return softmax(distances / two_sig_sq, zero_index=zero_index)


def find_optimal_sigmas(distances: np.ndarray, target_perplexity: int, max_iterations: int = 10000) -> np.ndarray:
    """
    For each row of distances matrix, find sigma that results
    in target perplexity for that role.
    (Implementation based around the following: https://nlml.github.io/in-raw-numpy/in-raw-numpy-t-sne/)
    """
    sigmas = []
    # For each row of the matrix (each point in our dataset)
    for i in range(distances.shape[0]):
        # Binary search over sigmas to achieve target perplexity
        lower_binary_search = 1e-20
        upper_binary_search = 1000.
        for _ in range(max_iterations):
            guess = (lower_binary_search + upper_binary_search) / 2.
            prob_mat = calc_prob_matrix(distances[i:i+1, :], np.array(guess), i)
            val = 2 ** -np.sum(prob_mat * np.log2(prob_mat), 1)

            if val > target_perplexity:
                upper_binary_search = guess
            else:
                lower_binary_search = guess
            if np.abs(val - target_perplexity) <= 1e-10:
                break

        correct_sigma = guess

        # Append the resulting sigma to our output array
        sigmas.append(correct_sigma)
    return np.array(sigmas)


def get_tsne_plot(df: pd.DataFrame,
                  perplexity: int = 20,
                  random_seed: int = 0,
                  num_iterations: int = 1000,
                  learning_rate: int = 500,
                  color_col: str = 'serial_number_id',
                  color_discrete_map: Optional[dict] = None,
                  momentum: int = 0) -> go.Figure:
    """
    Estimates a SNE model, and plots the visualization of the given data.

    More information about t-SNE visualizations and how to use them effectively can be found here:
    https://distill.pub/2016/misread-tsne/

    (Implementation based around the following: https://nlml.github.io/in-raw-numpy/in-raw-numpy-t-sne/)

    :param df: A Pandas Dataframe of the the data to be visualized (e.g. a recording attribute DataFrame)
    :param perplexity: A value which can be thought of as determining how many neighbors of a point
     will be used to update its position in the visualization.
    :param random_seed: Integer value to set the seed value for the random
    :param num_iterations: The number of iterations to train for
    :param learning_rate: The rate at which to update the values in the model
    :param color_col: The column name in the given dataframe (as ``merged_df``) that is used to color
     data points with.   This is used in combination with the color_discrete_map parameter
    :param color_discrete_map: A dictionary which maps the values given to color data points based on (see the
     color_col parameter description) to the colors that these data points should be
    :param momentum:  The momentum to be used when applying updates to the model
    :return: Plotly figure of the, low-dimensional representation of the given data
    """
    rng = np.random.RandomState(random_seed)

    X = df.loc[:, np.float64 == df.dtypes].dropna(axis='columns').to_numpy()

    # Standardize the data
    X = (X - X.mean()) / X.std()

    # Obtain matrix of joint probabilities p_ij
    euc_distances = neg_squared_euc_dists(X)
    calc_prob_mat = calc_prob_matrix(
        euc_distances,
        find_optimal_sigmas(euc_distances, perplexity))

    P = (calc_prob_mat + calc_prob_mat.T) / (2. * calc_prob_mat.shape[0])

    # Initialise our 2D representation
    Y = rng.normal(0., 0.0001, [X.shape[0], 2])

    # Initialise past values (used for momentum)
    if momentum:
        Y_m2 = Y.copy()
        Y_m1 = Y.copy()

    # Start gradient descent loop
    for i in range(num_iterations):

        # Get Q and distances (distances only used for t-SNE)
        distances = neg_squared_euc_dists(Y)
        inv_distances = np.power(1. - distances, -1)
        np.fill_diagonal(inv_distances, 0.)
        q_distances = inv_distances / np.sum(inv_distances)

        # Estimate gradients with respect to Y
        y_diffs = np.expand_dims(Y, 1) - np.expand_dims(Y, 0)
        grads = 4. * (np.expand_dims(P - q_distances, 2) * y_diffs * np.expand_dims(inv_distances, 2)).sum(1)

        # Update Y
        Y -= learning_rate * grads
        if momentum:  # Add momentum
            Y += momentum * (Y_m1 - Y_m2)
            # Update previous Y's for momentum
            Y_m2 = Y_m1.copy()
            Y_m1 = Y.copy()

    fig = px.scatter(
        x=Y[:, 0],
        y=Y[:, 1],
        color=df[color_col],
        color_discrete_map=color_discrete_map,
        hover_name=df.index,
    )

    fig.update_layout(
        xaxis_visible=False,
        yaxis_visible=False,
        showlegend=False,
        margin=dict(t=0, b=0, l=0, r=0),
    )

    return fig



from __future__ import annotations

from collections.abc import Container
import datetime
import typing
from typing import Optional
import warnings

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
from plotly import colors
from plotly.subplots import make_subplots
from scipy import spatial
import idelib.dataset

from endaq.calc import utils, psd
from endaq.ide import get_channel_table
from .utilities import determine_plotly_map_zoom, get_center_of_coordinates
from .dashboards import rolling_enveloped_dashboard

__all__ = [
    'multi_file_plot_attributes',
    'general_get_correlation_figure',
    'get_pure_numpy_2d_pca',
    'gen_map',
    'octave_spectrogram',
    'octave_psd_bar_plot',
    'rolling_min_max_envelope',
    'around_peak',
    'animate_quaternion',
    'spectrum_over_time',
    'pvss_on_4cp',
    'table_plot',
    'table_plot_from_ide'
]

DEFAULT_ATTRIBUTES_TO_PLOT_INDIVIDUALLY = np.array([
    'accelerationPeakFull', 'accelerationRMSFull', 'velocityRMSFull', 'psuedoVelocityPeakFull',
    'displacementRMSFull', 'gpsSpeedFull', 'gyroscopeRMSFull', 'microphonoeRMSFull',
    'temperatureMeanFull', 'pressureMeanFull'])


def multi_file_plot_attributes(multi_file_db: pd.DataFrame,
                               attribs_to_plot: np.ndarray = DEFAULT_ATTRIBUTES_TO_PLOT_INDIVIDUALLY,
                               recording_colors: Optional[Container] = None,
                               width_per_subplot: int = 400) -> go.Figure:
    """
    Creates a Plotly figure plotting all the desired attributes from the given DataFrame.

    :param multi_file_db: The Pandas DataFrame of data to plot attributes from
    :param attribs_to_plot: A numpy ndarray of strings with the names of the attributes to plot
    :param recording_colors: The colors to make each of the points (All will be the same color if None is given)
    :param width_per_subplot: The width to make every subplot
    :return: A Plotly figure of all the subplots desired to be plotted
    """
    if not isinstance(attribs_to_plot, np.ndarray):
        raise TypeError(
            "Instead of an ndarray given for 'attribs_to_plot', a variable of type %s was given" % str(type(attribs_to_plot)))

    if len(attribs_to_plot) == 0:
        raise ValueError("At least one value must be given for 'attribs_to_plot'")

    if not all(isinstance(row, str) for row in attribs_to_plot):
        raise TypeError("All rows to plot must be given as a string!")

    should_plot_row = np.array([not multi_file_db[r].isnull().all() for r in attribs_to_plot])

    attribs_to_plot = attribs_to_plot[should_plot_row]

    fig = make_subplots(
        rows=1,
        cols=len(attribs_to_plot),
        subplot_titles=attribs_to_plot,
    )

    for j, row_name in enumerate(attribs_to_plot):
        fig.add_trace(
            go.Scatter(
                x=multi_file_db['recording_ts'],
                y=multi_file_db[row_name],
                name=row_name,
                mode='markers',
                text=multi_file_db['serial_number_id'].values,
                marker_color=recording_colors,
            ),
            row=1,
            col=j + 1,
        )

    return fig.update_layout(width=len(attribs_to_plot)*width_per_subplot, showlegend=False)


def general_get_correlation_figure(merged_df: pd.DataFrame,
                                   color_col: str = 'serial_number_id',
                                   color_discrete_map: Optional[dict] = None,
                                   hover_names: Optional[Container] = None,
                                   characteristics_to_show_on_hover: list = [],
                                   starting_cols: Container = None) -> go.Figure:
    """
    A function to create a plot with two drop-down menus, each populated with a set of options corresponding to the
    scalar quantities contained in the given dataframe.   The data points will then be plotted with the X and Y axis
    corresponding to the selected attributes from the drop-down menu.

    :param merged_df: A Pandas DataFrame of data to use for producing the plot
    :param color_col: The column name in the given dataframe (as ``merged_df``) that is used to color
     data points with.   This is used in combination with the color_discrete_map parameter
    :param color_discrete_map: A dictionary which maps the values given to color data points based on (see the
     color_col parameter description) to the colors that these data points should be
    :param hover_names: The names of the points to display when they are hovered on
    :param characteristics_to_show_on_hover: The set of characteristics of the data to display when hovered over
    :param starting_cols: The two starting columns for the dropdown menus (will be the first two available
     if None is given)
    :return: The interactive Plotly figure
    """
    cols = [col for col, t in zip(merged_df.columns, merged_df.dtypes) if t != np.object]

    point_naming_characteristic = merged_df.index if hover_names is None else hover_names

    # This is not necessary, but usually produces easily discernible correlations or groupings of files/devices.
    # The hope is that when the initial plot has these characteristics, it will encourage
    # the exploration of this interactive plot.
    start_dropdown_indices = [0, 1]
    first_x_var = cols[0]
    first_y_var = cols[1]
    if starting_cols is not None and starting_cols[0] in cols and starting_cols[1] in cols:
        for j, col_name in enumerate(cols):
            if col_name == starting_cols[0]:
                first_x_var, start_dropdown_indices[0] = col_name, j
            if col_name == starting_cols[1]:
                first_y_var, start_dropdown_indices[1] = col_name, j

    # Create the scatter plot of the initially selected variables
    fig = px.scatter(
        merged_df,
        x=first_x_var,
        y=first_y_var,
        color=color_col,
        color_discrete_map=color_discrete_map,
        hover_name=point_naming_characteristic,
        hover_data=characteristics_to_show_on_hover,
        # width=800,
        # height=500,
    )

    # Create the drop-down menus which will be used to choose the desired file characteristics for comparison
    drop_downs = []
    for axis in ['x', 'y']:
        drop_downs.append([
            dict(
                method="update",
                args=[{axis: [merged_df[cols[k]]]},
                      {'%saxis.title.text' % axis: cols[k]},
                      # {'color': recording_colors},
                      ],
                label=cols[k]) for k in range(len(cols))
        ])

    # Sets up various aspects of the Plotly figure that is currently being produced.  This ranges from
    # aesthetic things, to setting the dropdown menus as part of the figure
    fig.update_layout(
        title_x=0.4,
        # width=800,
        # height=500,
        showlegend=False,
        updatemenus=[{
            'active': start_j,
            'buttons': drop_down,
            'x': 1.125,
            'y': y_height,
            'xanchor': 'left',
            'yanchor': 'top',
        } for drop_down, start_j, y_height in zip(drop_downs, start_dropdown_indices, [1, .85])])

    # Adds text labels for the two drop-down menus
    for axis, height in zip(['X', 'Y'], [1.05, .9]):
        fig.add_annotation(
            x=1.1,  # 1.185,
            y=height,
            xref='paper',
            yref='paper',
            showarrow=False,
            xanchor='left',
            yanchor='top',
            text="%s-Axis Measurement" % axis,
        )
    return fig


def get_pure_numpy_2d_pca(df: pd.DataFrame,
                          color_col: str = 'serial_number_id',
                          color_discrete_map: Optional[dict] = None,
                          ) -> go.Figure:
    """
    Get a Plotly figure of the 2D PCA for the given DataFrame.   This will have dropdown menus to select
    which components are being used for the X and Y axis.

    :param df: The dataframe of points to compute the PCA with
    :param color_col: The column name in the given dataframe (as ``df``) that is used to color
     data points with.   This is used in combination with the color_discrete_map parameter
    :param color_discrete_map: A dictionary which maps the values given to color data points based on (see the
     color_col parameter description) to the colors that these data points should be
    :return: A plotly figure as described in the main function description

    .. todo::
     - Add type checking statements to ensure the given dataframe contains enough values of the desired type
     - Add type checking statements to ensure the recording_colors given (if not None) are the proper length
    """

    # Drop all non-float64 type columns, and drop all columns with standard deviation of 0 because this will result
    # in division by 0
    X = df.loc[:, (df.std() != 0) & (np.float64 == df.dtypes)].dropna(axis='columns')

    # Standardize the values (so that mean of each variable is 0 and standard deviation is 1)
    X = (X - X.mean()) / X.std()

    # Get the shape of the data to compute PCA for
    n, m = X.shape

    # Compute covariance matrix
    covariance = np.dot(X.T, X) / (n - 1)

    # Eigen decomposition
    eigen_vals, eigen_vecs = np.linalg.eig(covariance)

    # Project X onto PC space
    X_pca = np.dot(X, eigen_vecs)

    # Create a new DataFrame with column names describing the fact that the values are the principal components
    pca_df = pd.DataFrame(X_pca)
    pca_df.columns = (pca_df.columns + 1).map(lambda x: "Component %d" % x)

    # Produce the Plotly figure for the points after being transfored with the PCA computed,
    # with dropdown menus to allow selection of what PCA components are being analyzed
    fig = general_get_correlation_figure(
        pca_df,
        color_col=df[color_col],
        color_discrete_map=color_discrete_map,
        characteristics_to_show_on_hover=[df.index],
        hover_names=df.index
    )

    return fig


def gen_map(df_map: pd.DataFrame,
            mapbox_access_token: str = None,
            lat: str = "Latitude",
            lon: str = "Longitude",
            color_by_column: str = None,
            filter_positive_color_vals: bool = False,
            hover_data: list[str] = [],
            size_max: float = 15.0,
            zoom_offset: float = -2.0
            ) -> go.Figure:
    """
    Plots GPS data on a map from a single recording, shading the points based on one of several characteristics
        (defaults to ground speed).
    
    :param df_map: The pandas dataframe containing the recording data.
    :param mapbox_access_token: The access token (or API key) needed to be able to plot against a map using Mapbox,
        `create a free account here <https://www.mapbox.com/pricing>`_

        * If no access token is provided, a `"stamen-terrain"` tile will be used,
            `see Plotly for more information <https://plotly.com/python/mapbox-layers/>`_
    :param lat: The dataframe column title to use for latitude
    :param lon: The dataframe column title to use for longitude
    :param color_by_column: The dataframe column title to color the plotted points by.
        If `None` is provided (the default), all the points will be the same color
    :param filter_positive_color_vals: A boolean variable, which will filter which points are plotted by
        if they have corresponding positive values, default is `False`
    :param hover_data: The list of dataframe column titles with data to display when the mouse hovers over a datapoint
    :param size_max: The size of the scatter points in the map, default 15
    :param zoom_offset: The offset to apply to the zoom, default -2, this is influenced by the final figure size

    Here is an example map of a drive from Boston Logan Airport to Mide Technology

    .. code:: python

        import endaq
        endaq.plot.utilities.set_theme()
        import pandas as pd

        # Get GPS Data
        gps = pd.read_csv('https://info.endaq.com/hubfs/data/mide-map-gps-data.csv')

        # Generate & Show Map
        fig = endaq.plot.gen_map(
            gps,
            lat="Latitude",
            lon="Longitude",
            color_by_column="Ground Speed",
            hover_data=["Date"]
        )
        fig.show()

    .. plotly::
        :fig-vars: fig

        import endaq
        endaq.plot.utilities.set_theme()
        import pandas as pd

        # Get GPS Data
        gps = pd.read_csv('https://info.endaq.com/hubfs/data/mide-map-gps-data.csv')

        # Generate & Show Map
        fig = endaq.plot.gen_map(
            gps,
            lat="Latitude",
            lon="Longitude",
            color_by_column="Ground Speed",
            hover_data=["Date"]
        )
        fig.show()
    """
    if filter_positive_color_vals and color_by_column is not None:
        df_map = df_map[df_map[color_by_column] > 0]
    
    zoom = determine_plotly_map_zoom(lats=df_map[lat], lons=df_map[lon])
    center = get_center_of_coordinates(lats=df_map[lat], lons=df_map[lon])
    
    px.set_mapbox_access_token(mapbox_access_token)
    
    fig = px.scatter_mapbox(
        df_map,
        lat=lat,
        lon=lon,
        color=color_by_column,
        hover_data=hover_data,
        size_max=size_max,
        zoom=zoom + zoom_offset,
        center=center,
    )

    if mapbox_access_token is None:
        fig.update_layout(mapbox_style="stamen-terrain")

    return fig.update_layout(margin={"r": 20, "t": 20, "l": 20, "b": 0})
    

def octave_spectrogram(df: pd.DataFrame, window: float, bins_per_octave: int = 3, freq_start: float = 20.0,
                       max_freq: float = float('inf'), db_scale: bool = True, log_scale_y_axis: bool = True
                       ) -> tuple[pd.DataFrame, go.Figure]:
    """
    Produces an octave spectrogram of the given data, this is a wrapper around 
    :py:func:`~endaq.calc.psd.rolling_psd()` and :py:func:`~endaq.plot.spectrum_over_time()`
    
    :param df: The dataframe of sensor data.  This must only have 1 column.
    :param window: The time window for each of the columns in the spectrogram
    :param bins_per_octave: The number of frequency bins per octave
    :param freq_start: The center of the first frequency bin
    :param max_freq: The maximum frequency to plot
    :param db_scale: If the spectrogram should be log-scaled for visibility (with 10*log10(x))
    :param log_scale_y_axis: If the y-axis of the plot should be log scaled
    :return: a tuple containing:
        - dataframe of the spectrogram data
        - the corresponding plotly figure
    """
    if len(df.columns) != 1:
        raise ValueError("The parameter 'df' must have only one column of data!")

    num_slices = int(len(df) * utils.sample_spacing(df) / window)

    df_psd = psd.rolling_psd(df, num_slices=num_slices, octave_bins=bins_per_octave,
                         fstart=freq_start)

    return df_psd, spectrum_over_time(df_psd, freq_max=max_freq, log_val=db_scale, log_freq=log_scale_y_axis)
    

def octave_psd_bar_plot(df: pd.DataFrame, bins_per_octave: int = 3, f_start: float = 20.0, yaxis_title: str = '',
                        log_scale_y_axis: bool = True) -> go.Figure:
    """
    Produces a bar plot of an octave psd.

    :param df: The dataframe of sensor data
    :param bins_per_octave: The number of frequency bins per octave
    :param f_start: The center of the first frequency bin
    :param yaxis_title: The text to label the y-axis
    :param log_scale_y_axis: If the y-axis should be log scaled
    """
    psd_df = psd.welch(df, 1, scaling='spectrum')

    octave_psd_df = psd.to_octave(
        psd_df,
        f_start, 
        bins_per_octave,
        agg='sum',
    )

    bar_widths = octave_psd_df.index * (2 ** (.5/bins_per_octave) - 2 ** (-.5/bins_per_octave))

    bar = go.Bar(
        x=octave_psd_df.index.values,
        y=octave_psd_df.values.squeeze(),
        width=bar_widths,
    )
    layout = go.Layout(
        yaxis={'title': yaxis_title},
        xaxis={'title': 'Frequency (Hz)'},
    )

    fig = go.Figure(data=bar, layout=layout)

    if log_scale_y_axis:
        fig.update_xaxes(type="log")

    return fig


def rolling_min_max_envelope(df: pd.DataFrame, desired_num_points: int = 2000, plot_as_bars: bool = True,
                             plot_title: str = "", opacity: float = 0.7,
                             colors_to_use: Optional[Container] = None) -> go.Figure:
    """
    A function to create a Plotly Figure to plot the data for each of the available data sub-channels, designed to
    reduce the number of points/data being plotted without minimizing the insight available from the plots.  It will
    plot either an envelope for rolling windows of the data (plotting the max and the min as line plots), or a bar based
    plot where the top of the bar (rectangle) is the highest value in the time window that bar spans, and the bottom of
    the bar is the lowest point in that time window (choosing between them is done with the `plot_as_bars` parameter).

    :param df: The dataframe of sub-channel data indexed by time stamps
    :param desired_num_points: The desired number of points to be plotted for each subchannel.  The number of points
     will be reduced from its original sampling rate by applying metrics (e.g. min, max) over sliding windows
     and then using that information to represent/visualize the data contained within the original data.  If less than
     the desired number of points are present, then a sliding window will NOT be used, and instead the points will be
     plotted as they were originally recorded  (also the subchannel will NOT be plotted as a bar based plot even if
     `plot_as_bars` was set to true).
    :param plot_as_bars: A boolean value indicating if the data should be visualized as a set of rectangles, where a
     shaded rectangle is used to represent the maximum and minimum values of the data during the time window
     covered by the rectangle.  These maximum and minimum values are visualized by the locations of the top and bottom
     edges of the rectangle respectively, unless the height of the rectangle would be 0, in which case a line segment
     will be displayed in its place.  If this parameter is `False`, two lines will be plotted for each
     of the sub-channels in the figure being created, creating an 'envelope' around the data.  An 'envelope' around the
     data consists of a line plotted for the maximum values contained in each of the time windows, and another line
     plotted for the minimum values.  Together these lines create a boundary which contains all the data points
     recorded in the originally recorded data.
    :param plot_title: The title for the plot
    :param opacity: The opacity to use for plotting bars/lines
    :param colors_to_use: An "array-like" object of strings containing colors to be cycled through for the sub-channels.
     If `None` is given (which is the default), then the `colorway` variable in Plotly's current theme/template will
     be used to color the data on each of the sub-channels uniquely, repeating from the start of the `colorway` if
     all colors have been used.
    :return: The Plotly Figure with the data plotted


    Here's an example plotting a dataset with over 6 million points per channel

    .. code:: python3

        import endaq
        endaq.plot.utilities.set_theme()
        import plotly.graph_objects as go

        # Get Accel, 6M datapoints per axis
        accel = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/ford_f150.ide',
            measurement_type='accel',
            time_mode='datetime')

        # Apply Highpass Filter
        accel = endaq.calc.filters.butterworth(accel, low_cutoff=2)

        # Generate Shaded Bar Plot of All Data
        fig = endaq.plot.rolling_min_max_envelope(accel)
        fig.show()

    .. plotly::
        :fig-vars: fig

        import endaq
        endaq.plot.utilities.set_theme()
        import plotly.graph_objects as go

        # Get Accel, 6M datapoints per axis
        accel = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/ford_f150.ide',
            measurement_type='accel',
            time_mode='datetime')

        # Apply Highpass Filter
        accel = endaq.calc.filters.butterworth(accel, low_cutoff=2)

        # Generate Shaded Bar Plot of All Data
        fig = endaq.plot.rolling_min_max_envelope(accel)
        fig.show()

    """

    return rolling_enveloped_dashboard(
        {plot_title: df},
        desired_num_points=desired_num_points,
        subplot_colors=colors_to_use,
        plot_as_bars=plot_as_bars,
        plot_full_single_channel=True,
        opacity=opacity
    )


def around_peak(df: pd.DataFrame, num: int = 1000, leading_ratio: float = 0.5):
    """
    A function to plot the data surrounding the largest peak (or valley) in the given data.
    The "peak" is defined by the point in the absolute value of the given data with the largest value.

    :param df: A dataframe indexed by time stamps
    :param num: The number of points to plot
    :param leading_ratio: The ratio of the data to be viewed that will come before the peak
    :return: A Plotly figure containing the plot
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"The `df` parmeter must be of type `pd.DataFrame` but was given type {type(df)} instead.")

    if not isinstance(num, int):
        raise TypeError(f"The `num` parameter must be an `int` type, but was given {type(num)}.")

    if not isinstance(leading_ratio, float):
        raise TypeError(f"The `leading_ratio` parameter must be a `float` type, but was given {type(leading_ratio)}.")

    if len(df) == 0:
        raise ValueError(f"The parameter `df` must have nonzero length, but has shape {df.shape} instead")

    if num < 3:
        raise ValueError(f"The `num` parameter must be at least 3, but {num} was given.")

    if leading_ratio < 0 or leading_ratio > 1:
        raise ValueError("The `leading_ratio` parameter must be a float value in the "
                         f"range [0,1], but was given {leading_ratio} instead.")

    max_i = df.abs().max(axis=1).reset_index(drop=True).idxmax()

    # These can go below and above the number of valid indices, but that can be ignored since
    # they'll only be used to slice the data in a way that is okay to go over/under
    window_start = max(0, max_i - int(num * leading_ratio))
    window_end = min(len(df), max_i + int(num * (1-leading_ratio)))

    return px.line(df.iloc[window_start: window_end])


def animate_quaternion(df, rate=6., scale=1.):
    """
    A function to animate orientation as a set of axis markers derived from quaternion data.

    :param df: A dataframe of quaternion data indexed by seconds.
               Columns must be 'X', 'Y', 'Z', and 'W'.  The order of the columns does not matter.
    :param rate: The number of frames per second to animate.
    :param scale: Time-scaling to adjust how quickly data is parsed.  Higher speeds will make the
                  file run faster, lower speeds will make the file run slower.
    :return: A Plotly figure containing the plot
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"The `df` parmeter must be of type `pd.DataFrame` but was given type {type(df)} instead.")
    elif len(df) == 0:
        raise ValueError(f"The parameter `df` must have nonzero length, but has shape {df.shape} instead")

    if not isinstance(rate, (int, float)):
        raise TypeError(f"The `rate` parameter must be an `int` or `float` type, but was given {type(rate)}.")
    elif rate <= 0:
        raise ValueError(f"The `rate` parameter must be a positive number, but was {rate}")

    if not isinstance(scale, (int, float)):
        raise TypeError(f"The `scale` parameter must be an `int` or `float` type, but was given {type(scale)}.")
    elif scale <= 0:
        raise ValueError(f"The `scale` parameter must be a positive number, but was {scale}")

    start = df.index[0]
    end = df.index[-1]

    fs = rate/scale

    t = np.arange(np.ceil((end - start)*fs))/fs + start

    r = spatial.transform.Rotation.from_quat(df[['X', 'Y', 'Z', 'W']])

    r_resample = spatial.transform.Slerp(df.index, r)(t)

    vals = []
    for _t, x, y, z in zip(
            t,
            r_resample.apply(([1, 0, 0])),
            r_resample.apply(([0, 1, 0])),
            r_resample.apply(([0, 0, 1])),
            ):

        vals.append({'time': _t, 'x': 0,    'y': 0,    'z': 0,    'channel': 'X'})
        vals.append({'time': _t, 'x': x[0], 'y': x[1], 'z': x[2], 'channel': 'X'})
        vals.append({'time': _t, 'x': 0,    'y': 0,    'z': 0,    'channel': 'Y'})
        vals.append({'time': _t, 'x': y[0], 'y': y[1], 'z': y[2], 'channel': 'Y'})
        vals.append({'time': _t, 'x': 0,    'y': 0,    'z': 0,    'channel': 'Z'})
        vals.append({'time': _t, 'x': z[0], 'y': z[1], 'z': z[2], 'channel': 'Z'})

    df_axes = pd.DataFrame(vals)

    fig = px.line_3d(
            df_axes,
            x='x',
            y='y',
            z='z',
            color='channel',
            animation_frame='time',
            animation_group='channel',
            ).update_layout(
            font_size=16,
            legend_title_text='',
            title_text='Quaternion 3D Animation',
            title_x=0.5,
            title_font_size=24,
            scene=dict(
                    xaxis=dict(range=[-1.1, 1.1], nticks=5),
                    yaxis=dict(range=[-1.1, 1.1], nticks=5),
                    zaxis=dict(range=[-1.1, 1.1], nticks=5),
                    aspectmode='cube',
                    ),
            )
    fig.layout.updatemenus[0].buttons[0].args[1]['frame']['duration'] = int(1000//rate)

    return fig


def spectrum_over_time(
        df: pd.DataFrame,
        plot_type: typing.Literal["Heatmap",
                                  "Surface",
                                  "Waterfall",
                                  "Animation",
                                  "Peak",
                                  "Lines"] = "Heatmap",
        var_column: str = 'variable',
        var_to_process: str = None,
        time_column: str = 'timestamp',
        freq_column: str = 'frequency (Hz)',
        val_column: str = 'value',
        freq_min: float = None,
        freq_max: float = None,
        log_freq: bool = False,
        log_val: bool = False,
        round_time: bool = True,
        round_freq: int = 2,
        waterfall_line_sequence: bool = True,
        zsmooth: str = "best",
        waterfall_line_color: str = '#EE7F27',
        min_median_max_line_color: str = '#6914F0',
) -> go.Figure:
    """
    Generate a 3D Plotly figure from a stacked spectrum to visualize how the frequency content changes over time
    
    :param df: the input dataframe with columns defining the frequency content, timestamps, values, and variables,
        see the following functions which provides outputs that would then be passed into this function as an input:

        *  :py:func:`~endaq.calc.fft.rolling_fft()`
        *  :py:func:`~endaq.calc.psd.rolling_psd()`
        *  :py:func:`~endaq.calc.shock.rolling_shock_spectrum()`
        *  :py:func:`~endaq.batch.GetDataBuilder.add_psd()`
        *  :py:func:`~endaq.batch.GetDataBuilder.add_pvss()`
    :param plot_type: the type of plot to display the spectrum, options are:

        *  `Heatmap`:  a 2D visualization with the color defining the z value (the default)
        *  `Surface`: similar to `Heatmap` but the z value is also projected "off the page"
        *  `Waterfall`: distinct lines are plotted per time slice in a 3D view
        *  `Animation`: a 2D display of the waterfall but the spectrum changes with animation frames
        *  `Peak`: per timestamp the peak frequency is determined and plotted against time
        *  `Lines`: the value in each frequency bin is plotted against time
    :param var_column: the column name in the dataframe that defines the different variables, default is `"variable"`
    :param var_to_process: the variable value in the `var_column` to filter the input df down to,
        if none is provided (the default) this function will filter to the first value
    :param time_column: the column name in the dataframe that defines the timestamps, default is `"timestamp"`
    :param freq_column: the column name in the dataframe that defines the frequency, default is `"frequency (Hz)"`
    :param val_column: the column name in the dataframe that defines the values, default is `"value"`
    :param freq_min: the minimum of the y axis (frequency) to include in the figure,
        default is None meaning it will display all content
    :param freq_max: the maximum of the y axis (frequency) to include in the figure,
        default is None meaning it will display all content
    :param log_freq: if `True` the frequency will be in a log scale, default is `False`
    :param log_val: if `True` the values will be in a log scale, default is `False`
    :param round_time: if `True` (default) the time values will be rounded to the nearest second for datetimes and
        hundredths of a second for floats
    :param round_freq: number of decimals to round the frequency bins to, default is 3
    :param waterfall_line_sequence: if `True` the waterfall line colors are defined with a color scale,
        if `False` all lines will have the same color, default is `True`
    :param zsmooth: the Plotly smooth algorithm to use in the heatmap, default is `"best"` which looks ideal but
        `"fast"` will be more responsive, or `False` will attempt no smoothing
    :param waterfall_line_color: the color to use for all lines in the Waterfall plot if `waterfall_line_sequence` is
        `False`
    :param min_median_max_line_color: the color to use for the min, max, and median lines in the Animation, if set to
        `None` these lines won't be added, default is `'#6914F0'`
    :return: a Plotly figure visualizing the spectrum over time


    Here's a few examples from a dataset recorded with an enDAQ sensor on a motorcycle
    as it revved the engine which resulted in changing frequency content
    
    .. code:: python

        import pandas as pd
        import endaq

        # Set Theme
        endaq.plot.utilities.set_theme()

        # Get Vibration Data
        df_vibe = pd.read_csv('https://info.endaq.com/hubfs/data/motorcycle-vibration-moving-frequency.csv', index_col=0)

        # Calculate a Rolling FFT
        fft = endaq.calc.fft.rolling_fft(df_vibe, num_slices=200, add_resultant=True)

        # Visualize the Rolling FFT as a Heatmap
        heatmap = endaq.plot.plots.spectrum_over_time(fft, plot_type='Heatmap', freq_max=200, var_to_process='Resultant')
        heatmap.show()

        # Plot the Peak Frequency vs Time
        peak = endaq.plot.plots.spectrum_over_time(fft, plot_type='Peak', freq_max=200, var_to_process='Resultant')
        peak.show()

        # Visualize as a Surface Plot
        surface = endaq.plot.plots.spectrum_over_time(fft, plot_type='Surface', freq_max=200, var_to_process='Resultant')
        surface.show()

        # Visualize as a Waterfall
        waterfall = endaq.plot.plots.spectrum_over_time(fft, plot_type='Waterfall', freq_max=200, var_to_process='Resultant')
        waterfall.show()

    .. plotly::
        :fig-vars: heatmap, peak, surface, waterfall

        import pandas as pd
        import endaq

        # Set Theme
        endaq.plot.utilities.set_theme()

        # Get Vibration Data
        df_vibe = pd.read_csv('https://info.endaq.com/hubfs/data/motorcycle-vibration-moving-frequency.csv', index_col=0)

        # Calculate a Rolling FFT
        fft = endaq.calc.fft.rolling_fft(df_vibe, num_slices=200, add_resultant=True)

        # Visualize the Rolling FFT as a Heatmap
        heatmap = endaq.plot.plots.spectrum_over_time(fft, plot_type='Heatmap', freq_max=200, var_to_process='Resultant')
        heatmap.show()

        # Plot the Peak Frequency vs Time
        peak = endaq.plot.plots.spectrum_over_time(fft, plot_type='Peak', freq_max=200, var_to_process='Resultant')
        peak.show()

        # Visualize as a Surface Plot
        surface = endaq.plot.plots.spectrum_over_time(fft, plot_type='Surface', freq_max=200, var_to_process='Resultant')
        surface.show()

        # Visualize as a Waterfall
        waterfall = endaq.plot.plots.spectrum_over_time(fft, plot_type='Waterfall', freq_max=200, var_to_process='Resultant')
        waterfall.show()

    Here's another few examples with a longer dataset with DatetimeIndex of a car engine during a morning commute

    .. code:: python

        import pandas as pd
        import endaq

        # Set Theme
        endaq.plot.utilities.set_theme()

        # Get a Longer Dataset with DatetimeIndex
        engine = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/data/Commute.ide', measurement_type='accel',
                                                   time_mode='datetime')

        # Compute PSD
        psd = endaq.calc.psd.rolling_psd(engine, num_slices=500, add_resultant=True, octave_bins=12, fstart=4)

        # Visualize as a Heatmap
        heatmap2 = endaq.plot.plots.spectrum_over_time(psd, plot_type='Heatmap', var_to_process='Resultant', zsmooth='best',
                                                       log_freq=True, log_val=True)
        heatmap2.show()

        # Visualize as an Animation
        animation = endaq.plot.plots.spectrum_over_time(psd, plot_type='Animation', var_to_process='Resultant',
                                                        log_freq=True, log_val=True)
        animation.show()

        # Use Rolling PSD to Calculate RMS in Certain Frequency Bins
        rms = endaq.calc.psd.rolling_psd(engine, num_slices=500, scaling='rms', add_resultant=True,
                                         freq_splits=[1, 20, 60, 300, 3000])

        # Plot the RMS per Frequency Bin Over Time
        lines = endaq.plot.plots.spectrum_over_time(rms, plot_type='Lines', log_val=False, var_to_process='Resultant')
        lines.show()

        # Compute Pseudo Velocity at Specific Times
        pvss = endaq.calc.shock.rolling_shock_spectrum(engine, slice_width=2.0, add_resultant=True,
                                                       mode='pvss', init_freq=4, damp=0.05,
                                                       index_values=pd.DatetimeIndex(['2016-08-02 12:07:15',
                                                                                      '2016-08-02 12:08:01',
                                                                                      '2016-08-02 12:10:06'], tz='UTC'))

        # Visualize as a Waterfall
        waterfall2 = endaq.plot.plots.spectrum_over_time(pvss, plot_type='Waterfall', log_freq=True, log_val=True,
                                                         var_to_process='Resultant', waterfall_line_sequence=False)
        waterfall2.show()

    .. plotly::
        :fig-vars: heatmap2, animation, lines, waterfall2

        import pandas as pd
        import endaq

        # Set Theme
        endaq.plot.utilities.set_theme()

        # Get a Longer Dataset with DatetimeIndex
        engine = endaq.ide.get_primary_sensor_data('https://info.endaq.com/hubfs/data/Commute.ide', measurement_type='accel',
                                                   time_mode='datetime')

        # Compute PSD
        psd = endaq.calc.psd.rolling_psd(engine, num_slices=500, add_resultant=True, octave_bins=12, fstart=4)

        # Visualize as a Heatmap
        heatmap2 = endaq.plot.plots.spectrum_over_time(psd, plot_type='Heatmap', var_to_process='Resultant', zsmooth='best',
                                                       log_freq=True, log_val=True)
        heatmap2.show()

        # Visualize as an Animation
        animation = endaq.plot.plots.spectrum_over_time(psd, plot_type='Animation', var_to_process='Resultant',
                                                        log_freq=True, log_val=True)
        animation.show()

        # Use Rolling PSD to Calculate RMS in Certain Frequency Bins
        rms = endaq.calc.psd.rolling_psd(engine, num_slices=500, scaling='rms', add_resultant=True,
                                         freq_splits=[1, 20, 60, 300, 3000])

        # Plot the RMS per Frequency Bin Over Time
        lines = endaq.plot.plots.spectrum_over_time(rms, plot_type='Lines', log_val=False, var_to_process='Resultant')
        lines.show()

        # Compute Pseudo Velocity at Specific Times
        pvss = endaq.calc.shock.rolling_shock_spectrum(engine, slice_width=2.0, add_resultant=True,
                                                       mode='pvss', init_freq=4, damp=0.05,
                                                       index_values=pd.DatetimeIndex(['2016-08-02 12:07:15',
                                                                                      '2016-08-02 12:08:01',
                                                                                      '2016-08-02 12:10:06'], tz='UTC'))

        # Visualize as a Waterfall
        waterfall2 = endaq.plot.plots.spectrum_over_time(pvss, plot_type='Waterfall', log_freq=True, log_val=True,
                                                         var_to_process='Resultant', waterfall_line_sequence=False)
        waterfall2.show()

    """
    df = df.copy()

    # Filter to one variable
    if var_to_process is None:
        var_to_process = df[var_column].unique()[0]
    df = df.loc[df[var_column] == var_to_process]

    # Round time
    if round_time:
        if isinstance(df[time_column].iloc[0], datetime.datetime):
            df[time_column] = df[time_column].round('s')
        else:
            df[time_column] = np.round(df[time_column].to_numpy(), 2)

    # Filter frequency
    df = df.loc[df[freq_column] > 0.0]
    if freq_max is not None:
        df = df.loc[df[freq_column] <= freq_max]
    if freq_min is not None:
        df = df.loc[df[freq_column] >= freq_min]

    # Remove 0s
    df = df.loc[df[val_column] > 0]

    # Check Length of Dataframe
    if len(df) > 100000:
        warnings.warn(
            "plot data is very large, may be unresponsive, suggest limiting frequency range and/or using less slices",
            RuntimeWarning,
        )

    # Create pivot table
    df_pivot = df.copy()
    first_step = df_pivot[freq_column].iloc[1] - df_pivot[freq_column].iloc[0]
    second_step = df_pivot[freq_column].iloc[2] - df_pivot[freq_column].iloc[1]
    if np.isclose(first_step, second_step):
        min_freq = np.round(df_pivot[freq_column].min(), round_freq)
        df_pivot[freq_column] = np.round(df_pivot[freq_column].to_numpy() / min_freq, 0) * min_freq
    df_pivot = df_pivot.pivot_table(columns=time_column, index=freq_column, values=val_column)

    # Heatmap & Surface
    if plot_type in ["Heatmap", "Surface"]:
        # Deal with Datetime
        x_type = float
        if isinstance(df_pivot.columns[0], datetime.datetime):
            if plot_type == "Surface":
                df_pivot.columns = (df_pivot.columns - df_pivot.columns[0]).total_seconds()
            else:
                x_type = str

        # Build Dictionary of Plot Data, Apply Log Scale If Needed
        data_dict = {
            'x': df_pivot.columns.astype(x_type),
            'y': df_pivot.index.astype(float),
            'z': df_pivot.to_numpy().astype(float),
            'connectgaps': True
        }
        if log_val:
            data_dict['z'] = np.log10(data_dict['z'])

        # Generate Figures
        if plot_type == "Heatmap":
            fig = go.Figure(data=go.Heatmap(data_dict, zsmooth=zsmooth)).update_layout(
                xaxis_title_text='Timestamp', yaxis_title_text='Frequency (Hz)')
            if log_freq:
                fig.update_layout(yaxis_type='log')
        else:
            fig = go.Figure(data=go.Surface(data_dict))
        fig.update_traces(showscale=False)

    # Waterfall
    elif plot_type == 'Waterfall':
        # Define Colors
        if waterfall_line_sequence:
            color_sequence = colors.sample_colorscale(
                [[0.0, '#6914F0'],
                 [0.2, '#3764FF'],
                 [0.4, '#2DB473'],
                 [0.6, '#FAC85F'],
                 [0.8, '#EE7F27'],
                 [1.0, '#D72D2D']], len(df[time_column].unique()))
        else:
            color_sequence = [waterfall_line_color]

        # Deal with Datetime
        df['label_column'] = df[time_column]
        if isinstance(df[time_column].iloc[0], datetime.datetime):
            df[time_column] = (df[time_column] - df[time_column].iloc[0]).dt.total_seconds()
            df['label_column'] = df['label_column'].dt.tz_convert(None).astype(str)

        # Generate Figure
        fig = px.line_3d(
            df,
            x=time_column,
            y=freq_column,
            z=val_column,
            color='label_column',
            color_discrete_sequence=color_sequence).update_layout(
            legend_orientation='v',
            legend_y=1,
            legend_title_text='Timestamps'
        )

    # Animation
    elif plot_type == 'Animation':
        # Deal with Datetime
        if isinstance(df[time_column].iloc[0], datetime.datetime):
            df[time_column] = df[time_column].dt.tz_convert(None).astype(str)

        # Generate Figure
        fig = px.line(
            df,
            animation_frame=time_column,
            x=freq_column,
            y=val_column,
            log_y=log_val,
            log_x=log_freq
        ).update_layout(
            showlegend=False,
            yaxis_title_text='',
            xaxis_title_text='Frequency (Hz)'
        )

        # Add min, max, median lines
        if min_median_max_line_color is not None:
            # Add Max
            df_t = df_pivot.max(axis=1).dropna()
            fig.add_trace(go.Scatter(
                x=df_t.index,
                y=df_t,
                mode='lines',
                line_color=min_median_max_line_color,
            ))

            # Add Min
            df_t = df_pivot.min(axis=1).dropna()
            fig.add_trace(go.Scatter(
                x=df_t.index,
                y=df_t,
                mode='lines',
                line_color=min_median_max_line_color,
            ))

            # Add Median
            df_t = df_pivot.median(axis=1).dropna()
            fig.add_trace(go.Scatter(
                x=df_t.index,
                y=df_t,
                mode='lines',
                line_color=min_median_max_line_color,
                line_dash='dash',
            ))

    # Peak Frequencies
    elif plot_type == 'Peak':
        fig = px.scatter(
            df_pivot.idxmax(),
            log_y=log_freq
        ).update_layout(
            showlegend=False,
            yaxis_title_text='Peak Frequency (Hz)',
            xaxis_title_text='Timestamp'
        )

    # Lines per Frequency Bin
    elif plot_type == 'Lines':
        df_pivot.index = np.round(df_pivot.index, 2)
        fig = px.line(
            df_pivot.T,
            log_y=log_val
        ).update_layout(
            legend_title_text='Frequency Bin (Hz)',
            yaxis_title_text='',
            xaxis_title_text='Timestamp',
            legend_orientation='v',
            legend_y=1,
        )

    else:
        raise ValueError(f"invalid plot type {plot_type}")

    # Add Labels to 3D plots
    if plot_type in ["Surface", "Waterfall"]:
        fig.update_scenes(
            aspectratio_x=2.0,
            aspectratio_y=1.0,
            aspectratio_z=0.3,
            xaxis_title_text='Timestamp',
            yaxis_title_text='Frequency (Hz)',
            zaxis_title_text=''
        )
        if log_freq:
            fig.update_scenes(yaxis_type='log')
        if log_val:
            fig.update_scenes(zaxis_type='log')
        fig.update_layout(scene_camera=dict(eye=dict(x=-1.5, y=-1.5, z=1)))

    return fig


def _log_ticks(start, stop, spacing):
    """Create an array of values that would be the equivalent of tick marks in a log scaled plot"""
    start = np.floor(np.log10(start))
    stop = np.ceil(np.log10(stop))
    if spacing == 'coarse':
        ones = np.array([1])
    elif spacing == 'medium':
        ones = np.array([1, 2, 5])
    else:
        ones = np.linspace(1, 9, 9)
    output = ones * (10 ** start)
    for i in np.arange(start + 1, stop, 1):
        output = np.concatenate((output,
                                 ones * (10 ** i)))
    return np.concatenate((output, np.array([10 ** stop])))


def _add_df(fig, df, line_color, units):
    """Add lines for each column in a dataframe to an existing figure"""
    for col in df.columns:
        fig.add_trace(go.Scatter(
            x=df.index,
            y=df[col].to_numpy(),
            mode="lines",
            line=dict(color=line_color, width=1, dash='solid'),
            showlegend=False,
            hoverinfo='skip',
            name=str(col) + units
        ))
    return fig


def pvss_on_4cp(
        df: pd.DataFrame,
        mode: typing.Literal["srs", "pvss"] = "srs",
        accel_units: str = "gravity",
        disp_units: str = "in",
        tick_spacing: typing.Literal["fine", "medium", "coarse"] = "medium",
        include_text: bool = True,
        size: typing.Optional[int] = None,
) -> go.Figure:
    """
    Given a shock response as a SRS or PVSS (see :py:func:`~endaq.calc.shock.shock_spectrum()`) return a plot of the
    pseudo velocity on a four coordinate plot (4CP or tripartite) that includes diagonal lines and hover labels to also
    signify the displacement and acceleration levels as a function of natural frequency.

    :param df: the input dataframe of shock response data, each column is plotted separately
    :param mode: the type of spectrum of the input dataframe, options are:

        *  `srs`:  default, shock response spectrum (SRS) which assumes has units of `accel_units`
        *  `pvss`: pseudo-velocity shock spectrum (PVSS) which assumes has units of `accel_units * s`
    :param accel_units: the units to display acceleration as, default is `"gravity"` which will be shortened to 'g'
        in labels, the unit conversion is handled using :py:func:`~endaq.calc.utils.convert_units()`
    :param disp_units: the units to display displacement as and velocity (divided by seconds), default is `"in"`,
        the unit conversion is handled using :py:func:`~endaq.calc.utils.convert_units()`
    :param tick_spacing: the spacing of each tick and corresponding diagonal line:

        *  `fine`:  order of magnitude, and linearly spaced ticks between each order of magnitude
        *  `medium`: default, order of magnitude, and then 2x and 5x between each order of magnitude, typical for Plotly,
        *  `coarse`: only the order of magnitude
    :param include_text: if `True` (default) add text labels to the diagonal lines
    :param size: the number of pixels to set the width and height of the figure to force it to be square,
        default is `None`
    :return: a Plotly figure of the PVSS on 4CP paper with hover information for acceleration and displacement


    Here's a few examples from a dataset recorded with an enDAQ sensor on a motorcycle
    as it revved the engine which resulted in changing frequency content

    .. code:: python3

        import endaq
        endaq.plot.utilities.set_theme()
        import pandas as pd

        # Get crash data
        df_crash = pd.read_csv('https://info.endaq.com/hubfs/data/motorcycle-crash.csv',index_col=0)

        # Calculate SRS
        srs = endaq.calc.shock.shock_spectrum(df_crash, mode='srs', damp=0.05)

        # Generate 4CP Plot
        imp = endaq.plot.plots.pvss_on_4cp(srs, disp_units='in')
        imp.show()

        # Change Plot Theme
        endaq.plot.utilities.set_theme('endaq_light')

        # Generate 4CP Plot with Different Units
        met = endaq.plot.plots.pvss_on_4cp(srs, disp_units='mm', tick_spacing='coarse', size=500)
        met.show()

    .. plotly::
        :fig-vars: imp, met

        import endaq
        endaq.plot.utilities.set_theme()
        import pandas as pd

        # Get crash data
        df_crash = pd.read_csv('https://info.endaq.com/hubfs/data/motorcycle-crash.csv',index_col=0)

        # Calculate SRS
        srs = endaq.calc.shock.shock_spectrum(df_crash, mode='srs', damp=0.05)

        # Generate 4CP Plot
        imp = endaq.plot.plots.pvss_on_4cp(srs, disp_units='in')
        imp.show()

        # Change Plot Theme
        endaq.plot.utilities.set_theme('endaq_light')

        # Generate 4CP Plot with Different Units
        met = endaq.plot.plots.pvss_on_4cp(srs, disp_units='mm', tick_spacing='coarse', size=500)
        met.show()

    """
    # Get and apply unit conversion
    accel_2_disp = utils.convert_units(units_in=accel_units, units_out=disp_units + '/s^2')
    df = df.copy() * accel_2_disp

    # Generate pseudo velocity data if srs is provided
    if mode == 'srs':
        df = df.div(2 * np.pi * df.index.to_series(), axis=0)

    # Generate acceleration & displacement dataframes
    accel = df.mul(2 * np.pi * df.index.to_series(), axis=0) / accel_2_disp
    disp = df.div(2 * np.pi * df.index.to_series(), axis=0)

    # Specify accel label
    accel_label = accel_units
    if accel_label == 'gravity':
        accel_label = 'g'

    # Get diagonal line information
    pvss_ticks = _log_ticks(
        start=10 ** np.floor(np.log10(df.min(axis=1).min())),
        stop=10 ** np.ceil(np.log10(df.max(axis=1).max())),
        spacing=tick_spacing
    )
    accel_ticks = _log_ticks(
        start=pvss_ticks[0] * df.index[0] * 2 * np.pi / accel_2_disp,
        stop=pvss_ticks[-1] * df.index[-1] * 2 * np.pi / accel_2_disp,
        spacing=tick_spacing
    )
    disp_ticks = _log_ticks(
        start=pvss_ticks[0] / (df.index[-1] * 2 * np.pi),
        stop=pvss_ticks[-1] / (df.index[0] * 2 * np.pi),
        spacing=tick_spacing
    )
    freqs = np.append(pvss_ticks[-1] / disp_ticks[0] / (2 * np.pi), df.index)
    freqs = np.append(freqs, pvss_ticks[-1] / disp_ticks[-1] / (2 * np.pi))
    accel_tick_df = pd.DataFrame(
        np.outer(1 / (freqs * 2 * np.pi), accel_ticks * accel_2_disp),
        index=freqs,
        columns=accel_ticks
    )
    disp_tick_df = pd.DataFrame(
        np.outer((freqs * 2 * np.pi), disp_ticks),
        index=freqs,
        columns=disp_ticks
    )

    # Generate figure & add shock spectrum data to plots
    fig = go.Figure()
    for c in df.columns:
        fig.add_trace(go.Scattergl(
            x=df.index.to_numpy(),
            y=df[c].to_numpy(),
            name=c,
            customdata=pd.DataFrame({
                    'Freq': df.index.to_numpy(),
                    'Accel': accel[c].to_numpy(),
                    'PV': df[c].to_numpy(),
                    'Disp': disp[c].to_numpy()}
            ),
            hovertemplate=c + '<br><br>' +
                          'Natural Frequency (Hz): %{customdata[0]:.2f}<br><br>' +
                          'Acceleration (' + accel_label + '): %{customdata[1]:.4E}<br>' +
                          'Pseudo-Velocity (' + disp_units + '/s): %{customdata[2]:.4E}<br>' +
                          'Displacement (' + disp_units + '): %{customdata[3]:.4E}<br><extra></extra>'
        ))

    # Add diagonal lines
    fig = _add_df(fig, accel_tick_df, fig.layout.template.layout.xaxis.gridcolor, accel_label)
    fig = _add_df(fig, disp_tick_df, fig.layout.template.layout.xaxis.gridcolor, disp_units)

    # Add text
    if include_text:
        dt = disp_tick_df.columns
        if tick_spacing == 'fine':
            dt = dt[np.remainder(np.log10(dt), 1) == 0]
        for disp in dt:
            disp_str = str(disp) + " " + disp_units
            if disp >= 1:
                disp_str = f"{disp:,.0f}" + " " + disp_units
            fig.add_annotation(
                x=np.log10(pvss_ticks[-1] / (disp * 2 * np.pi)),
                y=1,
                text=disp_str,
                yref='paper',
                xanchor='left',
                yanchor='bottom',
                textangle=-45,
                showarrow=False)
        at = accel_tick_df.columns
        if tick_spacing == 'fine':
            at = at[np.remainder(np.log10(at), 1) == 0]
        for accel in at:
            accel_str = str(accel) + " " + accel_label
            if accel >= 1:
                accel_str = f"{accel:,.0f}" + " " + accel_label
            fig.add_annotation(
                x=1,
                y=np.log10(accel / (df.index[-1] * 2 * np.pi) * accel_2_disp),
                text=accel_str,
                xref='paper',
                xanchor='left',
                yanchor='top',
                textangle=45,
                showarrow=False)

    return fig.update_layout(
        yaxis_type='log',
        xaxis_type='log',
        yaxis_range=np.log10([pvss_ticks[0], pvss_ticks[-1]]),
        xaxis_range=np.log10([df.index[0], df.index[-1]]),
        yaxis_title_text='Pseudo Velocity (' + disp_units + '/s)',
        xaxis_title_text='Natural Frequency (Hz)',
        width=size,
        height=size
    )


def table_plot(
        table: pd.DataFrame,
        num_round: typing.Optional[int] = 2,
        row_size: typing.Optional[int] = None,
        font_color: typing.Optional[str] = None,
        bg_color: typing.Optional[str] = None,
        line_color: typing.Optional[str] = None
) -> go.Figure:
    """
    Generate a Plotly figure from a Pandas dataframe that is styled consistent with the current Plotly template

    :param table: the input dataframe to generate the plot from
    :param num_round: the precision to round all numbers (if any) in `table` before displaying
    :param row_size: the size for each row, if `None` it will set this to 2 times the default Plotly template font size
    :param font_color: the color of all cell's text and the background color for the header, if `None` it will set this
        to the default Plotly font color
    :param bg_color: the color of all cell's background color and the text for the header, if `None` it will set this
        to the default Plotly background color
    :param line_color: the color for the lines in the table, if `None` it will set this
        to the default Plotly grid color
    :return: a Plotly table figure with all content from the dataframe


    Here's an example to generate a table from the metrics calculated with :py:func:`~endaq.calc.stats.shock_vibe_metrics()`

    .. code:: python3

        import endaq
        endaq.plot.utilities.set_theme('endaq_light')
        import pandas as pd

        # Get Acceleration Data
        accel = pd.read_csv('https://info.endaq.com/hubfs/Plots/bearing_data.csv', index_col=0)

        # Calculate Metrics
        metrics = endaq.calc.stats.shock_vibe_metrics(accel, include_resultant=False, freq_splits=[0, 65, 300, None])

        # Generate Plot to Show Metrics as a Table
        light_table = endaq.plot.table_plot(metrics)
        light_table.show()

        # Change Theme
        endaq.plot.utilities.set_theme('endaq')

        # Regenerate Plot to Show Metrics as a Table
        dark_table = endaq.plot.table_plot(metrics, num_round=4)
        dark_table.show()

    .. plotly::
        :fig-vars: light_table, dark_table

        import endaq
        endaq.plot.utilities.set_theme('endaq_light')
        import pandas as pd

        # Get Acceleration Data
        accel = pd.read_csv('https://info.endaq.com/hubfs/Plots/bearing_data.csv', index_col=0)

        # Calculate Metrics
        metrics = endaq.calc.stats.shock_vibe_metrics(accel, include_resultant=False, freq_splits=[0, 65, 300, None])

        # Generate Plot to Show Metrics as a Table
        light_table = endaq.plot.table_plot(metrics)
        light_table.show()

        # Change Theme
        endaq.plot.utilities.set_theme('endaq')

        # Regenerate Plot to Show Metrics as a Table
        dark_table = endaq.plot.table_plot(metrics, num_round=4)
        dark_table.show()

    """
    # Generate a Plotly Figure, then determine colors and sizes
    fig_template = go.Figure()
    if row_size is None:
        row_size = fig_template.layout.template.layout.font.size
        # Some Plotly templates don't specify this so we need to check if None again
        if row_size is None:
            row_size = 30
        else:
            row_size *= 2
    if bg_color is None:
        bg_color = fig_template.layout.template.layout.plot_bgcolor
    if font_color is None:
        font_color = fig_template.layout.template.layout.font.color
    if line_color is None:
        line_color = fig_template.layout.template.layout.xaxis.gridcolor

    # Generate Plot
    fig_table = go.Figure(data=[go.Table(
        header=dict(values=list(table.columns),
                    fill_color=font_color,
                    font_color=bg_color,
                    height=row_size),
        cells=dict(values=table.round(num_round).transpose().values.tolist(),
                   fill_color=bg_color,
                   height=row_size,
                   line_color=line_color),
    )
    ])
    return fig_table


def table_plot_from_ide(
        doc: idelib.dataset.Dataset = None,
        name: str = None,
        **kwargs
) -> go.Figure:
    """
    Generate a Plotly figure from a .IDE file using :py:func:`~endaq.plot.table_plot()` and
    :py:func:`~endaq.ide.get_channel_table()` while also displaying the device serial number, part number, and the
    date of the recording

    :param doc: A `idelib.dataset.Dataset`
    :param name: The plot title to add, if `None` then no title is added
    :param kwargs: Other parameters to pass directly to :py:func:`~endaq.plot.table_plot()`
    :return: a Plotly table figure with all content from the dataframe


    Here's an example to generate a table from the metrics calculated with :py:func:`~endaq.calc.stats.shock_vibe_metrics()`

    .. code:: python3

        import endaq
        endaq.plot.utilities.set_theme()

        doc = endaq.ide.get_doc('https://info.endaq.com/hubfs/data/All-Channels.ide')
        fig = endaq.plot.table_plot_from_ide(doc=doc, name='Example File')
        fig.show()

    .. plotly::
        :fig-vars: fig

        import endaq
        endaq.plot.utilities.set_theme()

        doc = endaq.ide.get_doc('https://info.endaq.com/hubfs/data/All-Channels.ide')
        fig = endaq.plot.table_plot_from_ide(doc=doc, name='Example File')
        fig.show()

    """
    # Get File Table
    file_table = get_channel_table(doc).data
    table = file_table[['name', 'type', 'units', 'samples', 'rate']].copy()

    # Get Figure
    fig_table = table_plot(table=table, **kwargs)

    # Add Title
    recorder_name = doc.recorderInfo['RecorderName']
    serial = doc.recorderInfo['RecorderSerial']
    part_no = doc.recorderInfo['PartNumber']
    date = pd.to_datetime(doc.lastUtcTime, unit='s')
    table_title = f'{recorder_name} (# {serial}, a {part_no})<br>Recorded on {date}'
    text_size = fig_table.layout.template.layout.font.size
    if text_size is None:
        text_size = 30
    else:
        text_size *= 2
    t_margin = text_size * 2 + 20
    if name is not None:
        table_title = name + '<br>' + table_title
        t_margin += text_size

    return fig_table.update_layout(
        title_text=table_title,
        title_y=.95, title_yanchor='top',
        margin=dict(l=20, r=20, t=t_margin, b=0))



from __future__ import annotations

from typing import Optional, Union, Tuple
import functools

import pandas as pd
import numpy as np
import scipy.signal

from endaq.calc import utils


def _get_filter_frequencies_type(low_cutoff, high_cutoff):
    """Get the filter type and cutoff frequency array."""
    cutoff_freqs: Union[float, Tuple[float, float]]
    filter_type: str

    if low_cutoff is not None and high_cutoff is not None:
        cutoff_freqs = (low_cutoff, high_cutoff)
        filter_type = "bandpass"
    elif low_cutoff is not None:
        cutoff_freqs = low_cutoff
        filter_type = "highpass"
    elif high_cutoff is not None:
        cutoff_freqs = high_cutoff
        filter_type = "lowpass"        
    else:
        return None, None
        
    return cutoff_freqs, filter_type        


def rolling_mean(
    df: pd.DataFrame,
    duration: float = 5.0
) -> pd.DataFrame:
    """
    Remove the rolling mean of an input time series dataframe

    :param df: the input data
    :param duration: the rolling window size in seconds to use
        - if `None` is given, the entire mean is removed 

    :return: a dataframe of the filtered data
    """
    if (duration is None):
        mean = df.mean()
    else:
        n = int(np.ceil(duration / utils.sample_spacing(df)) // 2 * 2 + 1)
        mean = df.rolling(n, min_periods=1, center=True).mean()    

    return df - mean   


def butterworth(
    df: pd.DataFrame,
    low_cutoff: Optional[float] = 1.0,
    high_cutoff: Optional[float] = None,
    half_order: int = 3,
    tukey_percent: float = 0.0,
) -> pd.DataFrame:
    """
    Apply a lowpass and/or a highpass Butterworth filter to an array.

    This function uses Butterworth filter designs, and implements the filter(s)
    as bi-directional digital biquad filters, split into second-order sections.

    :param df: the input data; cutoff frequencies are relative to the
        timestamps in `df.index`
    :param low_cutoff: the low-frequency cutoff, if any; frequencies below this
        value are rejected, and frequencies above this value are preserved
    :param high_cutoff: the high-frequency cutoff, if any; frequencies above
        this value are rejected, and frequencies below this value are preserved
    :param half_order: half of the order of the filter; higher orders provide
        more aggressive stopband reduction
    :param tukey_percent: the alpha parameter of a preconditioning Tukey filter;
        if 0 (default), no filter is applied
    :return: the filtered data

    .. seealso::

        - `SciPy Butterworth filter design <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html>`_
          Documentation for the butterworth filter design function.

        - `SciPy Tukey window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.tukey.html>`_
          Documentation for the Tukey window function used in preprocessing.
    """
    cutoff_freqs, filter_type = _get_filter_frequencies_type(low_cutoff, high_cutoff)

    if filter_type:
        dt = utils.sample_spacing(df)

        sos_coeffs = scipy.signal.butter(
            N=half_order,
            Wn=cutoff_freqs,
            btype=filter_type,
            fs=1 / dt,
            output="sos",
        )
        df = df.apply(functools.partial(scipy.signal.sosfiltfilt, sos_coeffs), axis=0)

    if tukey_percent > 0:
        tukey_window = scipy.signal.windows.tukey(len(df.index), alpha=tukey_percent)
        df = df.mul(tukey_window, axis="rows")

    return df


def bessel(
    df: pd.DataFrame,
    low_cutoff: Optional[float] = 1.0,
    high_cutoff: Optional[float] = None,
    half_order: int = 3,
    tukey_percent: float = 0.0,
    norm: typing.Literal["phase", "delay", "mag"] = "mag",
) -> pd.DataFrame:
    """
    Apply a lowpass and/or a highpass Bessel filter to an array.

    This function uses Bessel filter designs, and implements the filter(s)
    as bi-directional digital biquad filters, split into second-order sections.

    :param df: the input data; cutoff frequencies are relative to the
        timestamps in `df.index`
    :param low_cutoff: the low-frequency cutoff, if any; frequencies below this
        value are rejected, and frequencies above this value are preserved
    :param high_cutoff: the high-frequency cutoff, if any; frequencies above
        this value are rejected, and frequencies below this value are preserved
    :param half_order: half of the order of the filter; higher orders provide
        more aggressive stopband reduction
    :param tukey_percent: the alpha parameter of a preconditioning Tukey filter;
        if 0 (default), no filter is applied
    :param norm: how to normalize relative to the critical frequency:
            * `"phase"` - "phase-matched" case which is the default in SciPy & MATLAB
            * `"delay"` - the "natural" type obtained by solving Bessel polynomials
            * `"mag"` - gain magnitude is -3 dB at the cutoff frequency, default for this implementation to match Butterworth

    :return: the filtered data

    .. seealso::

        - `SciPy Bessel filter design <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.bessel.html>`_
          Documentation for the Bessel filter design function.

        - `SciPy Tukey window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.tukey.html>`_
          Documentation for the Tukey window function used in preprocessing.
    """
    cutoff_freqs, filter_type = _get_filter_frequencies_type(low_cutoff, high_cutoff)

    if filter_type:
        dt = utils.sample_spacing(df)

        sos_coeffs = scipy.signal.bessel(
            N=half_order,
            Wn=cutoff_freqs,
            btype=filter_type,
            fs=1 / dt,
            output="sos",
            norm=norm
        )
        df = df.apply(functools.partial(scipy.signal.sosfiltfilt, sos_coeffs), axis=0)

    if tukey_percent > 0:
        tukey_window = scipy.signal.windows.tukey(len(df.index), alpha=tukey_percent)
        df = df.mul(tukey_window, axis="rows")

    return df  


def cheby1(
    df: pd.DataFrame,
    low_cutoff: Optional[float] = 1.0,
    high_cutoff: Optional[float] = None,
    half_order: int = 3,
    tukey_percent: float = 0.0,
    rp: float = 3.0,
) -> pd.DataFrame:
    """
    Apply a lowpass and/or a highpass Chebyshev type I filter to an array.

    This function uses Chebyshev type I filter designs, and implements the filter(s)
    as bi-directional digital biquad filters, split into second-order sections.

    :param df: the input data; cutoff frequencies are relative to the
        timestamps in `df.index`
    :param low_cutoff: the low-frequency cutoff, if any; frequencies below this
        value are rejected, and frequencies above this value are preserved
    :param high_cutoff: the high-frequency cutoff, if any; frequencies above
        this value are rejected, and frequencies below this value are preserved
    :param half_order: half of the order of the filter; higher orders provide
        more aggressive stopband reduction
    :param tukey_percent: the alpha parameter of a preconditioning Tukey filter;
        if 0 (default), no filter is applied
    :param rp: the maximum ripple allowed in the passband, specified in decibels

    :return: the filtered data

    .. seealso::

        - `SciPy Cheby1 filter design <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.cheby1.html>`_
          Documentation for the Chebyshev type I filter design function.

        - `SciPy Tukey window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.tukey.html>`_
          Documentation for the Tukey window function used in preprocessing.
    """
    cutoff_freqs, filter_type = _get_filter_frequencies_type(low_cutoff, high_cutoff)

    if filter_type:
        dt = utils.sample_spacing(df)

        sos_coeffs = scipy.signal.cheby1(
            N=half_order,
            Wn=cutoff_freqs,
            btype=filter_type,
            fs=1 / dt,
            output="sos",
            rp=rp
        )
        df = df.apply(functools.partial(scipy.signal.sosfiltfilt, sos_coeffs), axis=0)

    if tukey_percent > 0:
        tukey_window = scipy.signal.windows.tukey(len(df.index), alpha=tukey_percent)
        df = df.mul(tukey_window, axis="rows")

    return df


def cheby2(
    df: pd.DataFrame,
    low_cutoff: Optional[float] = 1.0,
    high_cutoff: Optional[float] = None,
    half_order: int = 3,
    tukey_percent: float = 0.0,
    rs: float = 30.0,
) -> pd.DataFrame:
    """
    Apply a lowpass and/or a highpass Chebyshev type II filter to an array.

    This function uses Chebyshev type II filter designs, and implements the filter(s)
    as bi-directional digital biquad filters, split into second-order sections.

    :param df: the input data; cutoff frequencies are relative to the
        timestamps in `df.index`
    :param low_cutoff: the low-frequency cutoff, if any; frequencies below this
        value are rejected, and frequencies above this value are preserved
    :param high_cutoff: the high-frequency cutoff, if any; frequencies above
        this value are rejected, and frequencies below this value are preserved
    :param half_order: half of the order of the filter; higher orders provide
        more aggressive stopband reduction
    :param tukey_percent: the alpha parameter of a preconditioning Tukey filter;
        if 0 (default), no filter is applied
    :param rs: the minimum attenuation allowed in the stopband, specified in decibels

    :return: the filtered data

    .. seealso::

        - `SciPy Cheby2 filter design <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.cheby2.html>`_
          Documentation for the Chebyshev type II filter design function.

        - `SciPy Tukey window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.tukey.html>`_
          Documentation for the Tukey window function used in preprocessing.
    """
    cutoff_freqs, filter_type = _get_filter_frequencies_type(low_cutoff, high_cutoff)

    if filter_type:
        dt = utils.sample_spacing(df)

        sos_coeffs = scipy.signal.cheby2(
            N=half_order,
            Wn=cutoff_freqs,
            btype=filter_type,
            fs=1 / dt,
            output="sos",
            rs=rs
        )
        df = df.apply(functools.partial(scipy.signal.sosfiltfilt, sos_coeffs), axis=0)

    if tukey_percent > 0:
        tukey_window = scipy.signal.windows.tukey(len(df.index), alpha=tukey_percent)
        df = df.mul(tukey_window, axis="rows")

    return df           


def ellip(
    df: pd.DataFrame,
    low_cutoff: Optional[float] = 1.0,
    high_cutoff: Optional[float] = None,
    half_order: int = 3,
    tukey_percent: float = 0.0,
    rp: float = 3.0,
    rs: float = 30.0,
) -> pd.DataFrame:
    """
    Apply a lowpass and/or a highpass Elliptic filter to an array.

    This function uses Elliptic filter designs, and implements the filter(s)
    as bi-directional digital biquad filters, split into second-order sections.

    :param df: the input data; cutoff frequencies are relative to the
        timestamps in `df.index`
    :param low_cutoff: the low-frequency cutoff, if any; frequencies below this
        value are rejected, and frequencies above this value are preserved
    :param high_cutoff: the high-frequency cutoff, if any; frequencies above
        this value are rejected, and frequencies below this value are preserved
    :param half_order: half of the order of the filter; higher orders provide
        more aggressive stopband reduction
    :param tukey_percent: the alpha parameter of a preconditioning Tukey filter;
        if 0 (default), no filter is applied
    :param rp: the maximum ripple allowed in the passband, specified in decibels
    :param rs: the minimum attenuation allowed in the stopband, specified in decibels

    :return: the filtered data

    .. seealso::

        - `SciPy Ellip filter design <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.ellip.html>`_
          Documentation for the Elliptic filter design function.

        - `SciPy Tukey window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.tukey.html>`_
          Documentation for the Tukey window function used in preprocessing.
    """
    cutoff_freqs, filter_type = _get_filter_frequencies_type(low_cutoff, high_cutoff)

    if filter_type:
        dt = utils.sample_spacing(df)

        sos_coeffs = scipy.signal.ellip(
            N=half_order,
            Wn=cutoff_freqs,
            btype=filter_type,
            fs=1 / dt,
            output="sos",
            rs=rs,
            rp=rp
        )
        df = df.apply(functools.partial(scipy.signal.sosfiltfilt, sos_coeffs), axis=0)

    if tukey_percent > 0:
        tukey_window = scipy.signal.windows.tukey(len(df.index), alpha=tukey_percent)
        df = df.mul(tukey_window, axis="rows")

    return df            


def _fftnoise(f):
    """
    Generate time series noise for a given range of frequencies with random phase using ifft.
    """
    f = np.array(f, dtype='complex')
    Np = (len(f) - 1) // 2
    phases = np.random.rand(Np) * 2 * np.pi
    phases = np.cos(phases) + 1j * np.sin(phases)
    f[1:Np+1] *= phases
    f[-1:-1-Np:-1] = np.conj(f[1:Np+1])
    return np.fft.ifft(f).real


def band_limited_noise(
    min_freq: float = 0.0,
    max_freq: float = None,
    duration: float = 1.0,
    sample_rate: float = 1000.0,
    norm: typing.Literal["rms", "peak"] = "peak",
) -> pd.DataFrame:
    """
    Generate a time series with noise in a defined frequency range.

    :param min_freq: minimum frequency (Hz) where noise starts, default to 0
    :param max_freq: maximum frequency (Hz) where noise ends, default to 1/2 the sample rate
    :param duration: the duration of the time series returned, in seconds
    :param sample_rate: sample rate (Hz) of the time series
    :param norm: how to normalize the amplitude so that one of the following is equal to 1:
            * `"rms"` - root mean square
            * `"peak"` - peak value, default

    :return: a dataframe of the generated time series

    .. seealso::

        - `stack overflow post <https://stackoverflow.com/questions/33933842/how-to-generate-noise-in-frequency-range-with-numpy>`_
          that inspired this function and shared code we based this function on.
    
    """
    samples = int(duration * sample_rate)
    freqs = np.abs(np.fft.fftfreq(samples, 1/sample_rate))
    f = np.zeros(samples)

    if (max_freq is None):
        max_freq = sample_rate/2
    idx = np.where(np.logical_and(freqs>=min_freq, freqs<=max_freq))[0]
    f[idx] = 1

    df = pd.DataFrame(_fftnoise(f), 
                        index=pd.Series(np.arange(samples)/sample_rate, name='Time (s)'),
                        columns=['Noise from '+str(np.round(min_freq,1))+' to '+str(np.round(max_freq,1))+' Hz'])
    
    if (norm == 'rms'):
        df = df / np.mean(df**2)**0.5
    else:
        df = df / df.abs().max()

    return df




from __future__ import annotations

import typing
from typing import Optional
import warnings

import numpy as np
import pandas as pd
import scipy.fft

from endaq.calc import psd, utils


__all__ = ["aggregate_fft", "rolling_fft", "fft", "rfft", "dct", "dst"]


def aggregate_fft(df, **kwargs):
    """
    Calculate the FFT using :py:func:`scipy.signal.welch` with a specified frequency spacing.  The data returned is in the same
    units as the data input.

    :param df: The input data
    :param bin_width: The desired width of the resulting frequency bins, in Hz;
        defaults to 1 Hz
    :param kwargs: Other parameters to pass directly to :py:func:`scipy.signal.welch`
    :return: A periodogram of the input data in the same units as the input.

    .. seealso::

        - `SciPy Welch's method <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html>`_
          Documentation for the periodogram function wrapped internally.
    """
    kwargs.setdefault('scaling', 'unit')
    kwargs.setdefault('window', 'boxcar')
    return psd.welch(df, **kwargs)


def rolling_fft(
        df: pd.DataFrame,
        bin_width: float = 1.0,
        num_slices: int = 100,
        indexes: np.array = None,
        index_values: np.array = None,
        slice_width: float = None,
        add_resultant: bool = True,
        disable_warnings: bool = True,
        **kwargs,
) -> pd.DataFrame:
    """
    Compute FFTs for defined slices of a time series data set using :py:func:`~endaq.calc.fft.aggregate_fft()`
    
    :param df: the input dataframe with an index defining the time in seconds or datetime
    :param bin_width: the bin width or resolution in Hz for the FFT
    :param num_slices: the number of slices to split the time series into, default is 100,
        this is ignored if `indexes` is defined
    :param indexes: the center index locations (not value) of each slice to compute the FFT
    :param index_values: the index values of each peak event to quantify (slower but more intuitive than using `indexes`)
    :param slice_width: the time in seconds to center about each slice index,
        if none is provided it will calculate one based upon the number of slices
    :param add_resultant: if `True` the root sum of squares of each FFT column will
        also be computed
    :param disable_warnings: if `True` (default) it disables the warnings on the PSD length
    :param kwargs: Other parameters to pass directly to :py:func:`~endaq.calc.fft.aggregate_fft()`
    :return: a dataframe containing all the FFTs, stacked on each other

    See example use cases and syntax at :py:func:`~endaq.plot.spectrum_over_time()`
    which visualizes the output of this function in Heatmaps, Waterfall plots, 
    Surface plots, and Animations

    """
    use_spectrogram = True
    if (indexes is not None) | (index_values is not None):
        use_spectrogram = False

    indexes, slice_width, num, length = utils._rolling_slice_definitions(
        df,
        indexes=indexes,
        index_values=index_values,
        num_slices=num_slices,
        slice_width=slice_width
    )

    if use_spectrogram:
        fft_df = psd.spectrogram(
            df=df,
            num_slices=num_slices,
            bin_width=bin_width,
            scaling='unit',
            add_resultant=add_resultant,
            disable_warnings=disable_warnings,
            **kwargs
        )
    else:
        # Loop through and compute fft
        fft_df = pd.DataFrame()
        for i in indexes:
            window_start = max(0, i - num)
            window_end = min(length, i + num)
            with warnings.catch_warnings():
                if disable_warnings:
                    warnings.filterwarnings('ignore', '.*greater than.*')
                slice_fft = aggregate_fft(
                    df.iloc[window_start:window_end],
                    bin_width=bin_width,
                    **kwargs
                )
            if add_resultant:
                slice_fft['Resultant'] = slice_fft.pow(2).sum(axis=1).pow(0.5)

            slice_fft = slice_fft.reset_index().melt(id_vars=slice_fft.index.name)
            slice_fft['timestamp'] = df.index[i]
            fft_df = pd.concat([fft_df, slice_fft])

    return fft_df


def fft(
        df: pd.DataFrame,
        output: typing.Literal[None, "magnitude", "angle", "complex"] = None,
        nfft: Optional[int] = None,
        norm: typing.Literal[None, "unit", "forward", "ortho", "backward"] = None,
        optimize: bool = True,
    ) -> pd.DataFrame:
    """
    Perform the FFT of the data in ``df``, using SciPy's FFT method from :py:func:`scipy.fft.fft`.  If the in ``df`` is all real,
    then the output will be symmetrical between positive and negative frequencies, and it is instead recommended that
    you use the :py:func:`endaq.calc.fft.fft` method.

    :param df: The input data
    :param output: The type of the output of the FFT. Default is "magnitude".  "magnitude" will return the
                   absolute value of the FFT, "angle" will return the phase angle in radians, "complex" will return the
                   complex values of the FFT.
    :param nfft: Length of the transformed axis of the output. If nfft is smaller than the length of the
                 input, the input is cropped. If it is larger, the input is padded with zeros. If n is not given, the
                 length of the input along the axis specified by axis is used.
    :param norm: Normalization mode. Default is "unit", meaning a normalization of 2/n is applied on the
                 forward transform, and a normalization of 1/2 is applied on the ``ifft``. The "unit" normalization means
                 that the units of the FFT are the same as the units of the data put into it and that a sinusoid of
                 amplitude A will peak with amplitude A in the frequency domain.  “forward” instead applies a
                 normalization of 1/n on the forward transforms and no normalization is applied on the ``ifft``. “backward”
                 applies no normalization on the forward transform and 1/n on the backward. For ``norm="ortho"``, both
                 directions are scaled by `1/sqrt(n)`.
    :param optimize: If optimize is set to True, the length of the FFT will automatically be padded to a
                     length which can be calculated more quickly.  Default is True.
    :param kwargs: Further keywords passed to :py:func:`scipy.fft.fft`.  Note that the nfft parameter of this function is passed
                   to :py:func:`scipy.fft.fft` as ``n``.
    :return: The FFT of each channel in ``df``.

    .. seealso::

        - `SciPy FFT method <https://docs.scipy.org/doc/scipy/reference/reference/generated/scipy.fft.fft.html>`_
          Documentation for the FFT function wrapped internally.
    """
    if output is None:
        output_fun = np.abs
    elif output not in ["magnitude", "angle", "complex"]:
        raise ValueError(f'output must be one of None, "magnitude", "angle", or "complex".  Was {output}.')
    else:
        if output == "magnitude":
            output_fun = np.abs
        elif output == "angle":
            output_fun = np.angle
            norm = "forward"  # this prevents scaling when the angle is calculated
        elif output == "complex":
            output_fun = np.asarray
        else:
            raise Exception("impossible state reached in fft output type")

    if nfft is None:
        nfft = len(df)
    elif not isinstance(nfft, int):
        raise TypeError(f'nfft must be a positive integer, was of type {type(nfft)}')
    elif nfft <= 0:
        raise ValueError(f'nfft must be positive, was {nfft}')

    if optimize:
        nfft = scipy.fft.next_fast_len(nfft, False)

    scale = 1.
    if norm is None:
        norm = "forward"
        scale = 2.
    elif norm == "unit":
        norm = "forward"
        scale = 2.
    elif norm not in ["forward", "ortho", "backward"]:
        raise ValueError(f'norm must be one of None, "forward", "ortho", or "backward".  Was {norm}.')

    dt = utils.sample_spacing(df)

    data = {
            c: output_fun(scipy.fft.fftshift(scipy.fft.fft(
                df[c].to_numpy(),
                n=nfft,
                norm=norm,
            )))*scale
            for c in df.columns}
    index = pd.Index(scipy.fft.fftshift(scipy.fft.fftfreq(nfft, dt)), name='frequency (Hz)')

    return pd.DataFrame(
        data=data,
        columns=df.columns,
        index=index,
    )


def rfft(
        df: pd.DataFrame,
        output: typing.Literal[None, "magnitude", "angle", "complex"] = None,
        nfft: Optional[int] = None,
        norm: typing.Literal[None, "unit", "forward", "ortho", "backward"] = None,
        optimize: bool = True,
    ) -> pd.DataFrame:
    """
    Perform the real valued FFT of the data in ``df``, using SciPy's RFFT method from :py:func:`scipy.fft.rfft`.

    :param df: The input data
    :param output: The type of the output of the FFT. Default is "magnitude".  "magnitude" will return the
                   absolute value of the FFT, "angle" will return the phase angle in radians, "complex" will return the
                   complex values of the FFT.
    :param nfft: Length of the transformed axis of the output. If nfft is smaller than the length of the
                 input, the input is cropped. If it is larger, the input is padded with zeros. If n is not given, the
                 length of the input along the axis specified by axis is used.
    :param norm: Normalization mode. Default is "unit", meaning a normalization of 2/n is applied on the
                 forward transform, and a normalization of 1/2 is applied on the ``ifft``. The "unit" normalization means
                 that the units of the FFT are the same as the units of the data put into it and that a sinusoid of
                 amplitude A will peak with amplitude A in the frequency domain.  “forward” instead applies a
                 normalization of 1/n on the forward transforms and no normalization is applied on the ``ifft``. “backward”
                 applies no normalization on the forward transform and 1/n on the backward. For ``norm="ortho"``, both
                 directions are scaled by `1/sqrt(n)`.
    :param optimize: If optimize is set to True, the length of the FFT will automatically be padded to a
                     length which can be calculated more quickly.  Default is True.
    :param kwargs: Further keywords passed to :py:func:`scipy.fft.rfft`.  Note that the nfft parameter of this function is passed
                   to :py:func:`scipy.fft.rfft` as ``n``.
    :return: The RFFT of each channel in ``df``.

    .. seealso::

        - `SciPy RFFT method <https://docs.scipy.org/doc/scipy/reference/reference/generated/scipy.fft.rfft.html>`_
          Documentation for the RFFT function wrapped internally.
    """

    if output is None:
        output_fun = np.abs
    elif output not in ["magnitude", "angle", "complex"]:
        raise ValueError(f'output must be one of None, "magnitude", "angle", or "complex".  Was {output}.')
    else:
        if output == "magnitude":
            output_fun = np.abs
        elif output == "angle":
            output_fun = np.angle
            norm = "forward"  # this prevents scaling when the angle is calculated
        elif output == "complex":
            output_fun = np.asarray
        else:
            raise Exception("impossible state reached in fft output type")

    if nfft is None:
        nfft = len(df)
    elif not isinstance(nfft, int):
        raise TypeError(f'nfft must be a positive integer, was of type {type(nfft)}')
    elif nfft <= 0:
        raise ValueError(f'nfft must be positive, was {nfft}')

    if optimize:
        nfft = scipy.fft.next_fast_len(nfft, True)

    scale = 1.
    if norm is None:
        norm = "forward"
        scale = 2.
    elif norm == "unit":
        norm = "forward"
        scale = 2.
    elif norm not in ["forward", "ortho", "backward"]:
        raise ValueError(f'norm must be one of None, "forward", "ortho", or "backward".  Was {norm}.')

    dt = utils.sample_spacing(df)

    data = {
            c: output_fun(scipy.fft.rfft(
                df[c].to_numpy(),
                n=nfft,
                norm=norm,
            ))*scale
            for c in df.columns}
    index = pd.Index(scipy.fft.rfftfreq(nfft, dt), name='frequency (Hz)')

    return pd.DataFrame(
        data=data,
        columns=df.columns,
        index=index,
    )


def dct(
        df: pd.DataFrame,
        nfft: Optional[int] = None,
        norm: typing.Literal[None, "unit", "forward", "ortho", "backward"] = None,
        **kwargs,
    ) -> pd.DataFrame:
    """
    Calculate the DCT of the data in ``df``, using SciPy's DCT method from :py:func:`scipy.fft.dct`.

    :param df: the input data
    :param nfft: Length of the transformed axis of the output. If nfft is smaller than the length of the
                 input, the input is cropped. If it is larger, the input is padded with zeros. If n is not given, the
                 length of the input along the axis specified by axis is used.
    :param norm: Normalization mode. Default is "unit", meaning a normalization of 2/n is applied on the
                 forward transform, and a normalization of 1/2 is applied on the ``idct``. The "unit" normalization means
                 that the units of the FFT are the same as the units of the data put into it and that a sinusoid of
                 amplitude A will peak with amplitude A in the frequency domain.  “forward” instead applies a
                 normalization of 1/n on the forward transforms and no normalization is applied on the ``idct``. “backward”
                 applies no normalization on the forward transform and 1/n on the backward. For ``norm="ortho"``, both
                 directions are scaled by `1/sqrt(n)`.
    :param kwargs: Further keywords passed to :py:func:`scipy.fft.dct`.  Note that the nfft parameter of this function
                   is passed to :py:func:`scipy.fft.dct` as ``n``.
    :return: The DCT of each channel in ``df``.

    .. seealso::

        - `SciPy DCT method <https://docs.scipy.org/doc/scipy/reference/reference/generated/scipy.fft.dct.html>`_
          Documentation for the DCT function wrapped internally.
    """

    if nfft is None:
        nfft = len(df)
    elif not isinstance(nfft, int):
        raise TypeError(f'nfft must be a positive integer, was of type {type(nfft)}')
    elif nfft <= 0:
        raise ValueError(f'nfft must be positive, was {nfft}')

    scale = 1.

    if norm is None:
        norm = "forward"
        scale = 2.
    elif norm == "unit":
        norm = "forward"
        scale = 2.
    elif norm not in ["forward", "ortho", "backward"]:
        raise ValueError(f'norm must be one of None, "forward", "ortho", or "backward".  Was {norm}.')

    dt = utils.sample_spacing(df)
    fs = 1./dt

    return pd.DataFrame(
        data={
            c: scipy.fft.dst(
                df[c].to_numpy(),
                n=nfft,
                norm=norm,
                **kwargs,
            )*scale
            for c in df.columns},
        columns=df.columns,
        index=np.linspace(0, fs/2, nfft),
    )


def dst(
        df: pd.DataFrame,
        nfft: Optional[int] = None,
        norm: typing.Literal[None, "unit", "forward", "ortho", "backward"] = None,
        **kwargs,
    ) -> pd.DataFrame:
    """
    Calculate the DST of the data in ``df``, using SciPy's DST method from :py:func:`scipy.fft.dst`.

    :param df: the input data
    :param nfft: Length of the transformed axis of the output. If nfft is smaller than the length of the
                 input, the input is cropped. If it is larger, the input is padded with zeros. If n is not given, the
                 length of the input along the axis specified by axis is used.
    :param norm: Normalization mode. Default is "unit", meaning a normalization of 2/n is applied on the
                 forward transform, and a normalization of 1/2 is applied on the ``idst``. The "unit" normalization means
                 that the units of the FFT are the same as the units of the data put into it and that a sinusoid of
                 amplitude A will peak with amplitude A in the frequency domain.  “forward” instead applies a
                 normalization of 1/n on the forward transforms and no normalization is applied on the ``idst``. “backward”
                 applies no normalization on the forward transform and 1/n on the backward. For ``norm="ortho"``, both
                 directions are scaled by `1/sqrt(n)`.
    :param kwargs: Further keywords passed to :py:func:`scipy.fft.dst`.  Note that the nfft parameter of this function is passed
                   to :py:func:`scipy.fft.dst` as ``n``.
    :return: The DST of each channel in ``df``.

    .. seealso::

        - `SciPy DST method <https://docs.scipy.org/doc/scipy/reference/reference/generated/scipy.fft.dst.html>`_
          Documentation for the DST function wrapped internally.
    """

    if nfft is None:
        nfft = len(df)
    elif not isinstance(nfft, int):
        raise TypeError(f'nfft must be a positive integer, was of type {type(nfft)}')
    elif nfft <= 0:
        raise ValueError(f'nfft must be positive, was {nfft}')

    scale = 1.

    if norm is None:
        norm = "forward"
        scale = 2.
    elif norm == "unit":
        norm = "forward"
        scale = 2.
    elif norm not in ["forward", "ortho", "backward"]:
        raise ValueError(f'norm must be one of None, "forward", "ortho", or "backward".  Was {norm}.')

    dt = utils.sample_spacing(df)
    fs = 1./dt

    return pd.DataFrame(
        data={
            c: scipy.fft.dct(
                df[c].to_numpy(),
                n=nfft,
                norm=norm,
                **kwargs,
            )*scale
            for c in df.columns},
        columns=df.columns,
        index=np.linspace(0, fs/2, nfft),
    )



from __future__ import annotations

from collections.abc import Container
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.io as pio
from typing import Optional


def rolling_enveloped_dashboard(
    channel_df_dict: dict, desired_num_points: int = 250, num_rows: Optional[int] = None,
    num_cols: Optional[int] = 3, width_for_subplot_row: int = 400, height_for_subplot_row: int = 400,
    subplot_colors: Optional[Container] = None, min_points_to_plot: int = 1, plot_as_bars: bool = False,
    plot_full_single_channel: bool = False, opacity: float = 1.0, y_axis_bar_plot_padding: float = 0.06
) -> go.Figure:
    """
    A function to create a Plotly Figure with sub-plots for each of the available data sub-channels, designed to reduce
    the number of points/data being plotted without minimizing the insight available from the plots.  It will plot
    either an envelope for rolling windows of the data (plotting the max and the min as line plots), or a bar based
    plot where the top of the bar (rectangle) is the highest value in the time window that bar spans, and the bottom of
    the bar is the lowest point in that time window (choosing between them is done with the `plot_as_bars` parameter).


    :param channel_df_dict: A dictionary mapping channel names to Pandas DataFrames of that channels data
    :param desired_num_points: The desired number of points to be plotted in each subplot.  The number of points
     will be reduced from its original sampling rate by applying metrics (e.g. min, max) over sliding windows
     and then using that information to represent/visualize the data contained within the original data.  If less than
     the desired number of points are present, then a sliding window will NOT be used, and instead the points will be
     plotted as they were originally recorded  (also the subplot will NOT be plotted as a bar based plot even if
     `plot_as_bars` was set to true).
    :param num_rows: The number of columns of subplots to be created inside the Plotly figure. If `None` is given, (then
     `num_cols` must not be `None`), then this number will automatically be determined by what's needed.  If more rows
     are specified than are needed, the number of rows will be reduced to the minimum needed to contain all the subplots
    :param num_cols: The number of columns of subplots to be created inside the Plotly figure.  See the description of
     the `num_rows` parameter for more details on this parameter, and how the two interact.  This also follows the same
     approach to handling `None` when given
    :param width_for_subplot_row: The width of the area used for a single subplot (in pixels).
    :param height_for_subplot_row: The height of the area used for a single subplot (in pixels).
    :param subplot_colors: An "array-like" object of strings containing colors to be cycled through for the subplots.
     If `None` is given (which is the default), then the `colorway` variable in Plotly's current theme/template will
     be used to color the data on each of the subplots uniquely, repeating from the start of the `colorway` if
     all colors have been used.
    :param min_points_to_plot: The minimum number of data points required to be present to create a subplot for a
     channel/subchannel (NOT including `NaN` values).
    :param plot_as_bars: A boolean value indicating if the plot should be visualized as a set of rectangles, where a
     shaded rectangle is used to represent the maximum and minimum values of the data during the time window
     covered by the rectangle.  These maximum and minimum values are visualized by the locations of the top and bottom
     edges of the rectangle respectively, unless the height of the rectangle would be 0, in which case a line segment
     will be displayed in its place.  If this parameter is `False`, two lines will be plotted for each
     of the subplots in the figure being created, creating an "envelope" around the data.  An "envelope" around the
     data consists of a line plotted for the maximum values contained in each of the time windows, and another line
     plotted for the minimum values.  Together these lines create a boundary which contains all the data points
     recorded in the originally recorded data.
    :param plot_full_single_channel: If instead of a dashboard of subplots a single plot with multiple sub-channels
     should be created.  If this is True, only one (key, value) pair can be given for the `channel_df_dict` parameter
    :param opacity: The opacity to use for plotting bars/lines
    :param y_axis_bar_plot_padding: Due to some unknown reason the bar subplots aren't having their y axis ranges
     automatically scaled so this is the ratio of the total y-axis data range to pad both the top and bottom of the
     y axis with.  The default value is the one it appears Plotly uses as well.
    :return: The Plotly Figure containing the subplots of sensor data (the 'dashboard')


    Here's an example plotting a dashboard of all channels recorded on a commercial aircraft seat base

    .. code:: python3

        import endaq
        endaq.plot.utilities.set_theme('endaq_light')

        # Get IDE Object
        doc = endaq.ide.get_doc('https://info.endaq.com/hubfs/data/Seat-Base_21.ide')

        # Get All Data
        data = {doc.channels[ch].name: endaq.ide.to_pandas(doc.channels[ch], time_mode='datetime') for ch in doc.channels}

        # Get Dashboard of All Channels
        dashboard = endaq.plot.dashboards.rolling_enveloped_dashboard(
            data,
            num_rows=1,
            num_cols=None
        )
        dashboard.show()

    .. plotly::
        :fig-vars: dashboard

        import endaq
        endaq.plot.utilities.set_theme('endaq_light')

        # Get IDE Object
        doc = endaq.ide.get_doc('https://info.endaq.com/hubfs/data/Seat-Base_21.ide')

        # Get All Data
        data = {doc.channels[ch].name: endaq.ide.to_pandas(doc.channels[ch], time_mode='datetime') for ch in doc.channels}

        # Get Dashboard of All Channels
        dashboard = endaq.plot.dashboards.rolling_enveloped_dashboard(
            data,
            num_rows=1,
            num_cols=None
        )
        dashboard.show()

    """
    if not (num_rows is None or isinstance(num_rows, (int, np.integer))):
        raise TypeError(f"`num_rows` is of type `{type(num_rows)}`, which is not allowed.  " 
                        "`num_rows` can either be `None` or some type of integer.")
    elif not (num_cols is None or isinstance(num_cols, (int, np.integer))):
        raise TypeError(f"`num_cols` is of type `{type(num_cols)}`, which is not allowed.  "
                        "`num_cols` can either be `None` or some type of integer.")

    # I'm pretty sure the below is correct and it appears to work correctly, but it'd be nice if someone could
    # double check this logic

    # This removes any channels with less than `min_points_to_plot` data points per sub-channel, and removes any
    # sub-channels which have less than `min_points_to_plot` non-NaN data points
    channel_df_dict = {
        k: v.drop(columns=v.columns[v.notna().sum(axis=0) < min_points_to_plot])
            for (k, v) in channel_df_dict.items() if v.shape[0] >= min_points_to_plot}

    subplot_titles = [' '.join((k, col)) for (k, v) in channel_df_dict.items() for col in v.columns]

    if num_rows is None and num_cols is None:
        raise TypeError("Both `num_rows` and `num_columns` were given as `None`!  "
                        "A maximum of one of these two parameters may be given as None.")
    elif num_rows is None:
        num_rows = 1 + (len(subplot_titles) - 1) // num_cols
    elif num_cols is None:
        num_cols = 1 + (len(subplot_titles) - 1) // num_rows
    elif len(subplot_titles) > num_rows * num_cols:
        raise ValueError("The values given for `num_rows` and `num_columns` result in a maximum "
                        f"of {num_rows * num_cols} avaialable sub-plots, but {len(subplot_titles)} subplots need "
                        "to be plotted!   Try setting one of these variables to `None`, it will then "
                        "automatically be set to the optimal number of rows/columns.")
    else:
        num_rows = 1 + (len(subplot_titles) - 1) // num_cols
        num_cols = int(np.ceil(len(subplot_titles)/num_rows))

    if subplot_colors is None:
        colorway = pio.templates[pio.templates.default]['layout']['colorway']
    else:
        colorway = subplot_colors

    if plot_full_single_channel:
        if len(channel_df_dict) != 1:
            raise ValueError("The 'channel_df_dict' parameter must be length 1 when "
                             "'plot_full_single_channel' is set to true!")

        num_rows = 1
        num_cols = 1
        fig = go.Figure(layout_title_text=list(channel_df_dict.keys())[0])
    else:
        fig = make_subplots(
            rows=num_rows,
            cols=num_cols,
            subplot_titles=subplot_titles,
            figure=go.Figure(
                layout_height=height_for_subplot_row * num_rows,
                layout_width=width_for_subplot_row * num_cols,
            ),
        )

    # A counter to keep track of which subplot is currently being worked on
    subplot_num = 0

    # A dictionary to be used to modify the Plotly Figure layout all at once after all the elements have been added
    layout_changes_to_make = {}

    for channel_data in channel_df_dict.values():
        window = int(np.around((channel_data.shape[0]-1) / desired_num_points, decimals=0))

        # If it's going to be plotted as bars, force its time stamps to be uniformly spaced
        # so that the bars don't have any discontinuities in the X-axis
        if len(channel_data) >= desired_num_points and plot_as_bars:
            if isinstance(channel_data.index, pd.core.indexes.datetimes.DatetimeIndex):
                new_index = pd.date_range(
                    channel_data.index.values[0],
                    channel_data.index.values[-1],
                    periods=len(channel_data),
                )
            else:
                new_index = np.linspace(
                    channel_data.index.values[0],
                    channel_data.index.values[-1],
                    num=len(channel_data),
                )

            channel_data.set_index(new_index, inplace=True)


        # If a window size of 1 is determined, it sets the stride to 1 so we don't get an error as a result of the
        # 0 length stride
        stride = 1 if window == 1 else window - 1

        rolling_n = channel_data.rolling(window)
        min_max_tuple = (rolling_n.min()[::stride], rolling_n.max()[::stride])
        min_max_equal = min_max_tuple[0] == min_max_tuple[1]

        is_nan_min_max_mask = np.logical_not(
            np.logical_and(
                pd.isnull(min_max_tuple[0]),
                pd.isnull(min_max_tuple[1])))



        # Loop through each of the sub-channels, and their respective '0-height rectangle mask'
        for subchannel_name, cur_min_max_equal in min_max_equal[channel_data.columns].iteritems():

            traces = []
            cur_color = colorway[subplot_num % len(colorway)]
            cur_subchannel_non_nan_mask = is_nan_min_max_mask[subchannel_name]

            # If there are less data points than the desired number of points
            # to be plotted, just plot the data as a line plot
            if len(channel_data) < desired_num_points:
                not_nan_mask = np.logical_not(pd.isnull(channel_data[subchannel_name]))
                traces.append(
                    go.Scatter(
                        x=channel_data.index[not_nan_mask],
                        y=channel_data.loc[not_nan_mask, subchannel_name],
                        name=subchannel_name,
                        opacity=opacity,
                        line_color=cur_color,
                        showlegend=plot_full_single_channel,
                    )
                )
            # If it's going to plot the data with bars
            elif plot_as_bars:
                # If there are any 0-height rectangles
                if np.any(cur_min_max_equal):
                    equal_data_df = min_max_tuple[0].loc[cur_min_max_equal.values, subchannel_name]

                    # Half of the sampling period
                    half_dt = np.diff(cur_min_max_equal.index[[0, -1]])[0] / (2 * (len(cur_min_max_equal) - 1))

                    # Initialize the arrays we'll use for creating line segments where
                    # rectangles would have 0 width so it will end up formatted as follows
                    # (duplicate values for y since line segements are horizontal):
                    # x = [x1, x2, None, x3, x4, None, ...]
                    # y = [y12, y12, None, y34, y34, None, ...]
                    x_patch_line_segs = np.repeat(equal_data_df.index.values, 3)
                    y_patch_line_segs = np.repeat(equal_data_df.values, 3)

                    # All X axis values are the same, but these values are supposed to represent pairs of start and end
                    # times for line segments, so the time stamp is shifted half its duration backwards for the start
                    # time, and half its duration forward for the end time
                    x_patch_line_segs[::3] -= half_dt
                    x_patch_line_segs[1::3] += half_dt

                    # This is done every third value so that every two pairs of points is unconnected from eachother,
                    # since the (None, None) point will not connect to either the point before it nor behind it
                    x_patch_line_segs[2::3] = None
                    y_patch_line_segs[2::3] = None

                    traces.append(
                        go.Scatter(
                            x=x_patch_line_segs,
                            y=y_patch_line_segs,
                            name=subchannel_name,
                            opacity=opacity,
                            mode='lines',
                            line_color=cur_color,
                            showlegend=plot_full_single_channel,
                        )
                    )

                min_data_point = np.min(min_max_tuple[0][subchannel_name])
                max_data_point = np.max(min_max_tuple[1][subchannel_name])

                y_padding = (max_data_point - min_data_point) * y_axis_bar_plot_padding

                traces.append(
                    go.Bar(
                        x=min_max_tuple[0].index[cur_subchannel_non_nan_mask],
                        y=(min_max_tuple[1].loc[cur_subchannel_non_nan_mask, subchannel_name] -
                           min_max_tuple[0].loc[cur_subchannel_non_nan_mask, subchannel_name]),
                        marker_color=cur_color,
                        opacity=opacity,
                        marker_line_width=0,
                        base=min_max_tuple[0].loc[cur_subchannel_non_nan_mask, subchannel_name],
                        showlegend=plot_full_single_channel,
                        name=subchannel_name,
                    )
                )

                # Adds a (key, value) pair to the dict for setting this subplot's Y-axis display range (applied later)
                min_y_range = min_data_point - y_padding
                max_y_range = max_data_point + y_padding
                y_axis_id = f'yaxis{1 + subplot_num}_range'
                if plot_full_single_channel:
                    y_axis_id = 'yaxis_range'
                    if layout_changes_to_make:
                        min_y_range = min(min_y_range, layout_changes_to_make[y_axis_id][0])
                        max_y_range = max(max_y_range, layout_changes_to_make[y_axis_id][1])

                layout_changes_to_make[y_axis_id] = [min_y_range, max_y_range]
            else:
                for cur_df in min_max_tuple:
                    traces.append(
                        go.Scatter(
                            x=cur_df.index[cur_subchannel_non_nan_mask],
                            y=cur_df.loc[cur_subchannel_non_nan_mask, subchannel_name],
                            name=subchannel_name,
                            opacity=opacity,
                            line_color=cur_color,
                            showlegend=plot_full_single_channel,
                        )
                    )

            # Add the traces created for the current subchannel of data to the plotly figure
            if plot_full_single_channel:
                fig.add_traces(traces)
            else:
                fig.add_traces(
                    traces,
                    rows=1 + subplot_num // num_cols,
                    cols=1 + subplot_num % num_cols,
                )

            subplot_num += 1

    fig.update_layout(
        **layout_changes_to_make,
        bargap=0,
        barmode='overlay'
    )
    return fig


def rolling_metric_dashboard(channel_df_dict: dict, desired_num_points: int = 250, num_rows: Optional[int] = None,
             num_cols: Optional[int] = 3, rolling_metrics_to_plot: tuple = ('mean', 'std'),
             metric_colors: Optional[Container] = None, width_for_subplot_row: int = 400,
             height_for_subplot_row: int = 400) -> go.Figure:
    """
    A function to create a dashboard of subplots of the given data, plotting a set of rolling metrics.

    :param channel_df_dict: A dictionary mapping channel names to Pandas DataFrames of that channels data
    :param desired_num_points:  The desired number of points to be plotted in each subplot.  The number of points
     will be reduced from its original sampling rate by applying metrics (e.g. min, max) over sliding windows
     and then using that information to represent/visualize the data contained within the original data.  If less than
     the desired number of points are present, then a sliding window will NOT be used, and instead the points will be
     plotted as they were originally recorded  (also the subplot will NOT be plotted as a bar based plot even if
     `plot_as_bars` was set to true).
    :param num_rows: The number of columns of subplots to be created inside the Plotly figure. If `None` is given, (then
     `num_cols` must not be `None`), then this number will automatically be determined by what's needed.  If more rows
     are specified than are needed, the number of rows will be reduced to the minimum needed to contain all the subplots
    :param num_cols: The number of columns of subplots to be created inside the Plotly figure.  See the description of
     the `num_rows` parameter for more details on this parameter, and how the two interact.  This also follows the same
     approach to handling `None` when given
    :param rolling_metrics_to_plot: A tuple of strings which indicate what rolling metrics to plot for each subchannel.
     The options are ``['mean', 'std', 'absolute max', 'rms']`` which correspond to the mean, standard deviation, maximum of
     the absolute value, and root-mean-square.
    :param metric_colors: An "array-like" object of strings containing colors to be cycled through for the metrics.
     If `None` is given (which is the default), then the `colorway` variable in Plotly's current theme/template will
     be used to color the metric data, repeating from the start of the `colorway` if all colors have been used.
     The first value corresponds to the color if not enough points of data exist for a rolling metric,
     and the others correspond to the metric in `rolling_metrics_to_plot` in the same order they are given
    :param width_for_subplot_row: The width of the area used for a single subplot (in pixels).
    :param height_for_subplot_row: The height of the area used for a single subplot (in pixels).
    :return: The Plotly Figure containing the subplots of sensor data (the 'dashboard')
    """
    if len(rolling_metrics_to_plot) == 0:
        raise ValueError("At least one rolling metric must be specified in `rolling_metrics_to_plot`!")

    subplot_titles = [' '.join((k, col)) for (k, v) in channel_df_dict.items() for col in v.columns]

    if num_rows is None and num_cols is None:
        raise TypeError("Both `num_rows` and `num_columns` were given as `None`!  "
                        "A maximum of one of these two parameters may be given as None.")
    elif num_rows is None:
        num_rows = 1 + (len(subplot_titles) - 1) // num_cols
    elif num_cols is None:
        num_cols = 1 + (len(subplot_titles) - 1) // num_rows
    elif len(subplot_titles) > num_rows * num_cols:
        raise ValueError("The values given for `num_rows` and `num_columns` result in a maximum "
                         f"of {num_rows * num_cols} avaialable sub-plots, but {len(subplot_titles)} subplots need "
                         "to be plotted!   Try setting one of these variables to `None`, it will then "
                         "automatically be set to the optimal number of rows/columns.")
    else:
        num_rows = 1 + (len(subplot_titles) - 1) // num_cols
        num_cols = int(np.ceil(len(subplot_titles)/num_rows))

    if metric_colors is None:
        colorway = pio.templates[pio.templates.default]['layout']['colorway']
    else:
        colorway = metric_colors
    fig = make_subplots(rows=num_rows, cols=num_cols, subplot_titles=subplot_titles)
    subplot_num = 0
    for channel_name, channel_data in channel_df_dict.items():
        n = int(channel_data.shape[0] / desired_num_points)

        if n > 0:
            time = channel_data.rolling(n).mean().iloc[::n].index
        else:
            time = channel_data.index

        for c_i, (subchannel_name, subchannel_data) in enumerate(channel_data.iteritems()):
            for metric in rolling_metrics_to_plot:
                if n == 0:
                    data = subchannel_data.values
                    name = subchannel_name
                    color = colorway[0]
                elif metric == 'mean':
                    data = subchannel_data.rolling(n).mean().iloc[::n]
                    name = 'Mean'
                    color = colorway[1]
                elif metric == 'std':
                    data = subchannel_data.rolling(n).std().iloc[::n]
                    name = 'STD Dev'
                    color = colorway[2]
                elif metric == 'absolute max':
                    data = subchannel_data.abs().rolling(n).max().iloc[::n]
                    name = 'Max'
                    color = colorway[3]
                elif metric == 'rms':
                    data = subchannel_data.pow(2).rolling(n).mean().apply(np.sqrt, raw=True).iloc[::n]
                    name = 'RMS'
                    color = colorway[4]
                else:
                    raise ValueError(f"metric given to `rolling_metrics_to_plot` is not valid!  Was given {metric}"
                                     " which is not in the allowed options of ['mean', 'std', 'absolute max', 'rms']")
                fig.add_trace(
                    go.Scatter(
                        x=time,
                        y=data,
                        name=name,
                        line=dict(color=color)),
                    row=1 + subplot_num // num_cols,
                    col=1 + subplot_num % num_cols,
                )

                if n == 0:
                    break

            subplot_num += 1

    return fig.update_layout(
        width=num_cols * width_for_subplot_row,
        height=num_rows * height_for_subplot_row,
        showlegend=False,
    )






if __name__ == '__main__':
    import endaq.ide
    from utilities import set_theme
    set_theme()

    file_urls = ['https://info.endaq.com/hubfs/data/surgical-instrument.ide',
                 'https://info.endaq.com/hubfs/data/97c3990f-Drive-Home_70-1616632444.ide',
                 'https://info.endaq.com/hubfs/data/High-Drop.ide',
                 'https://info.endaq.com/hubfs/data/HiTest-Shock.ide',
                 'https://info.endaq.com/hubfs/data/Drive-Home_01.ide',
                 'https://info.endaq.com/hubfs/data/Tower-of-Terror.ide',
                 'https://info.endaq.com/hubfs/data/Punching-Bag.ide',
                 'https://info.endaq.com/hubfs/data/Gun-Stock.ide',
                 'https://info.endaq.com/hubfs/data/Seat-Base_21.ide',
                 'https://info.endaq.com/hubfs/data/Seat-Top_09.ide',
                 'https://info.endaq.com/hubfs/data/Bolted.ide',
                 'https://info.endaq.com/hubfs/data/Motorcycle-Car-Crash.ide',
                 'https://info.endaq.com/hubfs/data/train-passing.ide',
                 'https://info.endaq.com/hubfs/data/baseball.ide',
                 'https://info.endaq.com/hubfs/data/Clean-Room-VC.ide',
                 'https://info.endaq.com/hubfs/data/enDAQ_Cropped.ide',
                 'https://info.endaq.com/hubfs/data/Drive-Home_07.ide',
                 'https://info.endaq.com/hubfs/data/ford_f150.ide',
                 'https://info.endaq.com/hubfs/data/Drive-Home.ide',
                 'https://info.endaq.com/hubfs/data/Mining-Data.ide',
                 'https://info.endaq.com/hubfs/data/Mide-Airport-Drive-Lexus-Hybrid-Dash-W8.ide']

    for j in [4]:
        doc = endaq.ide.get_doc(file_urls[j])
        table = endaq.ide.get_channel_table(doc)

        # (IMPORTANT NOTE) The use of this as a dictionary is dependent on it maintaining being 'insertion ordered',
        # which is a thing in Python 3.7 (may have existed in a different way in python 3.6, but I'm not sure)
        CHANNEL_DFS = {
            doc.channels[ch].name: endaq.ide.to_pandas(doc.channels[ch], time_mode='seconds') for ch in doc.channels}

        SINGLE_CHANNEL = r'40g DC Acceleration'
        JUST_ACCEL_DFS = {SINGLE_CHANNEL: CHANNEL_DFS[SINGLE_CHANNEL]}

        # Examples
        rolling_enveloped_dashboard(
            JUST_ACCEL_DFS,
            plot_full_single_channel=True,
        ).show()

        rolling_enveloped_dashboard(
            JUST_ACCEL_DFS,
            plot_as_bars=True,
            plot_full_single_channel=True,
        ).show()

        rolling_enveloped_dashboard(
            CHANNEL_DFS,
            desired_num_points=100,
            min_points_to_plot=10,
            plot_as_bars=True,
            height_for_subplot_row=600,
            width_for_subplot_row=600,
            num_cols=2,
            num_rows=None,
        ).show()

        rolling_enveloped_dashboard(
            CHANNEL_DFS,
            desired_num_points=1000,
            num_rows=2,
            num_cols=9999999,
            width_for_subplot_row=250,
            height_for_subplot_row=250,
        ).show()

        rolling_enveloped_dashboard(
            CHANNEL_DFS,
            desired_num_points=1000,
            num_rows=9999999,
            num_cols=2,
            width_for_subplot_row=250,
            height_for_subplot_row=250,
        ).show()

        rolling_enveloped_dashboard(
            CHANNEL_DFS,
            desired_num_points=1000,
            min_points_to_plot=10,
            plot_as_bars=True,
        ).show()

        rolling_enveloped_dashboard(
            CHANNEL_DFS,
            desired_num_points=1000,
            min_points_to_plot=10,
            num_rows=999999,
            num_cols=4,
        ).show()

        rolling_metric_dashboard(
            CHANNEL_DFS,
            rolling_metrics_to_plot=('mean', 'absolute max', 'std'),
        ).show()

        print(str(j) + " done!")
